import os
import logging
import asyncio

import unittest

from src.modules.speech.tts.cosy_voice2_tts import CosyVoiceTTS
from src.common.factory import EngineFactory, EngineClass
from src.common.logger import Logger
from src.common.session import Session
from src.common.utils.wav import save_audio_to_file
from src.common.types import SessionCtx, CosyVoiceTTSArgs, MODELS_DIR

r"""
python -m unittest test.modules.speech.tts.test_cosy_voice.TestCosyVoiceTTS.test_get_voices

# cosyvoice default use sft infer
LOG_LEVEL=INFO TTS_TAG=tts_cosy_voice MODELS_DIR=/content/achatbot/models/FunAudioLLM/CosyVoice-300M-SFT \
  python -m unittest test.modules.speech.tts.test_cosy_voice.TestCosyVoiceTTS.test_synthesize

# cosyvoice test_synthesize with instruct infer
LOG_LEVEL=INFO TTS_TAG=tts_cosy_voice MODELS_DIR=/content/achatbot/models/FunAudioLLM/CosyVoice-300M-Instruct \
  TTS_TEXT="在面对挑战时，他展现了非凡的<strong>勇气</strong>与<strong>智慧</strong>。" \
  INSTRUCT_TEXT="Theo \'Crimson\', is a fiery, passionate rebel leader. Fights with fervor for justice, but struggles with impulsiveness." \
  python -m unittest test.modules.speech.tts.test_cosy_voice.TestCosyVoiceTTS.test_synthesize

# cosyvoice test_synthesize with zero shot infer
# zero_shot usage, <|zh|><|en|><|jp|><|yue|><|ko|> for Chinese/English/Japanese/Cantonese/Korean
LOG_LEVEL=INFO TTS_TAG=tts_cosy_voice MODELS_DIR=/content/achatbot/models/FunAudioLLM/CosyVoice-300M \
  TTS_TEXT="您好啊，我是🤖机器人，很高兴认识您~" \
  REFERENCE_AUDIO_PATH=/content/achatbot/test/audio_files/asr_example_zh.wav \
  REFERENCE_TEXT="欢迎大家来体验达摩院推出的语音识别模型。" \
  python -m unittest test.modules.speech.tts.test_cosy_voice.TestCosyVoiceTTS.test_synthesize

# cosyvoice test_synthesize with cross_lingual infer
LOG_LEVEL=INFO TTS_TAG=tts_cosy_voice MODELS_DIR=/content/achatbot/models/FunAudioLLM/CosyVoice-300M \
  TTS_TEXT="<|en|>And then later on, fully acquiring that company. So keeping management in line, interest in line with the asset that\'s coming into the family is a reason why sometimes we don\'t buy the whole thing." \
  REFERENCE_AUDIO_PATH=/content/achatbot/test/audio_files/asr_example_zh.wav \
  python -m unittest test.modules.speech.tts.test_cosy_voice.TestCosyVoiceTTS.test_synthesize

# cosyvoice test_synthesize with voice convert infer
LOG_LEVEL=INFO TTS_TAG=tts_cosy_voice MODELS_DIR=/content/achatbot/models/FunAudioLLM/CosyVoice-300M \
  SRC_AUDIO_PATH=/content/achatbot/deps/CosyVoice/asset/cross_lingual_prompt.wav \
  REFERENCE_AUDIO_PATH=/content/achatbot/test/audio_files/asr_example_zh.wav \
  python -m unittest test.modules.speech.tts.test_cosy_voice.TestCosyVoiceTTS.test_synthesize


------

# cosyvoice2 no spk, so need reference audio
LOG_LEVEL=INFO TTS_TAG=tts_cosy_voice2 MODELS_DIR=/content/achatbot/models/FunAudioLLM/CosyVoice2-0.5B \
  REFERENCE_AUDIO_PATH=/content/achatbot/records/test_tts_cosy_voice.wav \
  python -m unittest test.modules.speech.tts.test_cosy_voice.TestCosyVoiceTTS.test_get_voices

# cosyvoice2 no spk, so need reference audio
LOG_LEVEL=INFO TTS_TAG=tts_cosy_voice2 MODELS_DIR=/content/achatbot/models/FunAudioLLM/CosyVoice2-0.5B \
  REFERENCE_AUDIO_PATH=/content/achatbot/records/test_tts_cosy_voice.wav \
  python -m unittest test.modules.speech.tts.test_cosy_voice.TestCosyVoiceTTS.test_get_voices

# cosyvoice2 test_synthesize with instruct2 infer
LOG_LEVEL=INFO TTS_TAG=tts_cosy_voice2 MODELS_DIR=/content/achatbot/models/FunAudioLLM/CosyVoice2-0.5B \
  REFERENCE_AUDIO_PATH=/content/achatbot/records/test_tts_cosy_voice.wav \
  INSTRUCT_TEXT="用四川话说这句话" \
  python -m unittest test.modules.speech.tts.test_cosy_voice.TestCosyVoiceTTS.test_synthesize

# cosyvoice2 test_synthesize with zero shot infer
LOG_LEVEL=INFO TTS_TAG=tts_cosy_voice2 MODELS_DIR=/content/achatbot/models/FunAudioLLM/CosyVoice2-0.5B \
  TTS_TEXT="您好啊，我是🤖机器人，很高兴认识您~" \
  REFERENCE_AUDIO_PATH=/content/achatbot/records/test_tts_cosy_voice.wav \
  REFERENCE_TEXT="你好，我是机器人, hello, test.modules.speech.tts.test_gtts.TestGTTS.test_synthesize" \
  python -m unittest test.modules.speech.tts.test_cosy_voice.TestCosyVoiceTTS.test_synthesize

# cosyvoice2 test_synthesize with cross lingual infer and use fine grained control, 
# for supported control, check cosyvoice/tokenizer/tokenizer.py#L248
LOG_LEVEL=DEBUG TTS_TAG=tts_cosy_voice2 MODELS_DIR=/content/achatbot/models/FunAudioLLM/CosyVoice2-0.5B \
  TTS_TEXT="在他讲述那个荒诞故事的过程中，他突然[laughter]停下来，因为他自己也被逗笑了[laughter]。" \
  REFERENCE_AUDIO_PATH=/content/achatbot/records/test_tts_cosy_voice.wav \
  python -m unittest test.modules.speech.tts.test_cosy_voice.TestCosyVoiceTTS.test_synthesize
"""


class TestCosyVoiceTTS(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.tts_tag = os.getenv("TTS_TAG", "tts_cosy_voice")
        cls.tts_text = os.getenv(
            "TTS_TEXT",
            "你好，我是机器人, hello, test.modules.speech.tts.test_gtts.TestGTTS.test_synthesize",
        )
        model_dir = os.path.join(MODELS_DIR, "CosyVoice-300M-SFT")
        cls.model_dir = os.getenv("MODELS_DIR", model_dir)
        Logger.init(os.getenv("LOG_LEVEL", "debug").upper(), is_file=False)

    @classmethod
    def tearDownClass(cls):
        pass

    def setUp(self):
        kwargs = CosyVoiceTTSArgs(
            model_dir=self.model_dir,
            reference_text=os.getenv("REFERENCE_TEXT", ""),
            reference_audio_path=os.getenv("REFERENCE_AUDIO_PATH", ""),
            src_audio_path=os.getenv("SRC_AUDIO_PATH", ""),
            instruct_text=os.getenv("INSTRUCT_TEXT", ""),
            spk_id=os.getenv(
                "SPK_ID", ""
            ),  # for cosyvoice sft/struct inference, cosyvoice2 don't use it
        ).__dict__
        self.tts: CosyVoiceTTS = EngineFactory.get_engine_by_tag(
            EngineClass, self.tts_tag, **kwargs
        )
        self.session = Session(**SessionCtx("test_tts_client_id").__dict__)
        self.pyaudio_instance = None
        self.audio_stream = None

    def tearDown(self):
        self.audio_stream and self.audio_stream.stop_stream()
        self.audio_stream and self.audio_stream.close()
        self.pyaudio_instance and self.pyaudio_instance.terminate()

    def test_get_voices(self):
        voices = self.tts.get_voices()
        self.assertGreaterEqual(len(voices), 0)
        print(voices)

    def test_synthesize(self):
        stream_info = self.tts.get_stream_info()
        print(stream_info)
        self.session.ctx.state["tts_text"] = self.tts_text
        print(self.session.ctx)
        iter = self.tts.synthesize_sync(self.session)
        res = bytearray()
        for i, chunk in enumerate(iter):
            print(i, len(chunk))
            res.extend(chunk)
        path = asyncio.run(
            save_audio_to_file(
                res,
                f"test_{self.tts.TAG}.wav",
                sample_rate=stream_info["rate"],
                sample_width=stream_info["sample_width"],
                channles=stream_info["channels"],
            )
        )
        print(path)

    def test_synthesize_speak(self):
        import pyaudio

        stream_info = self.tts.get_stream_info()
        self.pyaudio_instance = pyaudio.PyAudio()
        self.audio_stream = self.pyaudio_instance.open(
            format=stream_info["format"],
            channels=stream_info["channels"],
            rate=stream_info["rate"],
            output_device_index=None,
            output=True,
        )

        self.session.ctx.state["tts_text"] = self.tts_text
        print(self.session.ctx)
        iter = self.tts.synthesize_sync(self.session)
        sub_chunk_size = 1024
        for i, chunk in enumerate(iter):
            print(f"get {i} chunk {len(chunk)}")
            self.assertGreaterEqual(len(chunk), 0)
            if len(chunk) / sub_chunk_size < 100:
                self.audio_stream.write(chunk)
                continue
            for i in range(0, len(chunk), sub_chunk_size):
                sub_chunk = chunk[i : i + sub_chunk_size]
                self.audio_stream.write(sub_chunk)



import gradio as gr
import torch
from janus.janusflow.models import MultiModalityCausalLM, VLChatProcessor
from PIL import Image
from diffusers.models import AutoencoderKL
import numpy as np

cuda_device = 'cuda' if torch.cuda.is_available() else 'cpu'

# Load model and processor
model_path = "deepseek-ai/JanusFlow-1.3B"
vl_chat_processor = VLChatProcessor.from_pretrained(model_path)
tokenizer = vl_chat_processor.tokenizer

vl_gpt = MultiModalityCausalLM.from_pretrained(model_path)
vl_gpt = vl_gpt.to(torch.bfloat16).to(cuda_device).eval()

# remember to use bfloat16 dtype, this vae doesn't work with fp16
vae = AutoencoderKL.from_pretrained("stabilityai/sdxl-vae")
vae = vae.to(torch.bfloat16).to(cuda_device).eval()

# Multimodal Understanding function
@torch.inference_mode()
# Multimodal Understanding function
def multimodal_understanding(image, question, seed, top_p, temperature):
    # Clear CUDA cache before generating
    torch.cuda.empty_cache()
    
    # set seed
    torch.manual_seed(seed)
    np.random.seed(seed)
    torch.cuda.manual_seed(seed)
    
    conversation = [
        {
            "role": "User",
            "content": f"<image_placeholder>\n{question}",
            "images": [image],
        },
        {"role": "Assistant", "content": ""},
    ]
    
    pil_images = [Image.fromarray(image)]
    prepare_inputs = vl_chat_processor(
        conversations=conversation, images=pil_images, force_batchify=True
    ).to(cuda_device, dtype=torch.bfloat16 if torch.cuda.is_available() else torch.float16)
    
    
    inputs_embeds = vl_gpt.prepare_inputs_embeds(**prepare_inputs)
    
    outputs = vl_gpt.language_model.generate(
        inputs_embeds=inputs_embeds,
        attention_mask=prepare_inputs.attention_mask,
        pad_token_id=tokenizer.eos_token_id,
        bos_token_id=tokenizer.bos_token_id,
        eos_token_id=tokenizer.eos_token_id,
        max_new_tokens=512,
        do_sample=False if temperature == 0 else True,
        use_cache=True,
        temperature=temperature,
        top_p=top_p,
    )
    
    answer = tokenizer.decode(outputs[0].cpu().tolist(), skip_special_tokens=True)

    return answer


@torch.inference_mode()
def generate(
    input_ids,
    cfg_weight: float = 2.0,
    num_inference_steps: int = 30
):
    # we generate 5 images at a time, *2 for CFG
    tokens = torch.stack([input_ids] * 10).cuda()
    tokens[5:, 1:] = vl_chat_processor.pad_id
    inputs_embeds = vl_gpt.language_model.get_input_embeddings()(tokens)
    print(inputs_embeds.shape)

    # we remove the last <bog> token and replace it with t_emb later
    inputs_embeds = inputs_embeds[:, :-1, :] 
    
    # generate with rectified flow ode
    # step 1: encode with vision_gen_enc
    z = torch.randn((5, 4, 48, 48), dtype=torch.bfloat16).cuda()
    
    dt = 1.0 / num_inference_steps
    dt = torch.zeros_like(z).cuda().to(torch.bfloat16) + dt
    
    # step 2: run ode
    attention_mask = torch.ones((10, inputs_embeds.shape[1]+577)).to(vl_gpt.device)
    attention_mask[5:, 1:inputs_embeds.shape[1]] = 0
    attention_mask = attention_mask.int()
    for step in range(num_inference_steps):
        # prepare inputs for the llm
        z_input = torch.cat([z, z], dim=0) # for cfg
        t = step / num_inference_steps * 1000.
        t = torch.tensor([t] * z_input.shape[0]).to(dt)
        z_enc = vl_gpt.vision_gen_enc_model(z_input, t)
        z_emb, t_emb, hs = z_enc[0], z_enc[1], z_enc[2]
        z_emb = z_emb.view(z_emb.shape[0], z_emb.shape[1], -1).permute(0, 2, 1)
        z_emb = vl_gpt.vision_gen_enc_aligner(z_emb)
        llm_emb = torch.cat([inputs_embeds, t_emb.unsqueeze(1), z_emb], dim=1)

        # input to the llm
        # we apply attention mask for CFG: 1 for tokens that are not masked, 0 for tokens that are masked.
        if step == 0:
            outputs = vl_gpt.language_model.model(inputs_embeds=llm_emb, 
                                             use_cache=True, 
                                             attention_mask=attention_mask,
                                             past_key_values=None)
            past_key_values = []
            for kv_cache in past_key_values:
                k, v = kv_cache[0], kv_cache[1]
                past_key_values.append((k[:, :, :inputs_embeds.shape[1], :], v[:, :, :inputs_embeds.shape[1], :]))
            past_key_values = tuple(past_key_values)
        else:
            outputs = vl_gpt.language_model.model(inputs_embeds=llm_emb, 
                                             use_cache=True, 
                                             attention_mask=attention_mask,
                                             past_key_values=past_key_values)
        hidden_states = outputs.last_hidden_state
        
        # transform hidden_states back to v
        hidden_states = vl_gpt.vision_gen_dec_aligner(vl_gpt.vision_gen_dec_aligner_norm(hidden_states[:, -576:, :]))
        hidden_states = hidden_states.reshape(z_emb.shape[0], 24, 24, 768).permute(0, 3, 1, 2)
        v = vl_gpt.vision_gen_dec_model(hidden_states, hs, t_emb)
        v_cond, v_uncond = torch.chunk(v, 2)
        v = cfg_weight * v_cond - (cfg_weight-1.) * v_uncond
        z = z + dt * v
        
    # step 3: decode with vision_gen_dec and sdxl vae
    decoded_image = vae.decode(z / vae.config.scaling_factor).sample
    
    images = decoded_image.float().clip_(-1., 1.).permute(0,2,3,1).cpu().numpy()
    images = ((images+1) / 2. * 255).astype(np.uint8)
    
    return images
    
def unpack(dec, width, height, parallel_size=5):
    dec = dec.to(torch.float32).cpu().numpy().transpose(0, 2, 3, 1)
    dec = np.clip((dec + 1) / 2 * 255, 0, 255)

    visual_img = np.zeros((parallel_size, width, height, 3), dtype=np.uint8)
    visual_img[:, :, :] = dec

    return visual_img


@torch.inference_mode()
def generate_image(prompt,
                   seed=None,
                   guidance=5,
                   num_inference_steps=30):
    # Clear CUDA cache and avoid tracking gradients
    torch.cuda.empty_cache()
    # Set the seed for reproducible results
    if seed is not None:
        torch.manual_seed(seed)
        torch.cuda.manual_seed(seed)
        np.random.seed(seed)
    
    with torch.no_grad():
        messages = [{'role': 'User', 'content': prompt},
                    {'role': 'Assistant', 'content': ''}]
        text = vl_chat_processor.apply_sft_template_for_multi_turn_prompts(conversations=messages,
                                                                   sft_format=vl_chat_processor.sft_format,
                                                                   system_prompt='')
        text = text + vl_chat_processor.image_start_tag
        input_ids = torch.LongTensor(tokenizer.encode(text))
        images = generate(input_ids,
                                   cfg_weight=guidance,
                                   num_inference_steps=num_inference_steps)
        return [Image.fromarray(images[i]).resize((1024, 1024), Image.LANCZOS) for i in range(images.shape[0])]

        

# Gradio interface
with gr.Blocks() as demo:
    gr.Markdown(value="# Multimodal Understanding")
    # with gr.Row():
    with gr.Row():
        image_input = gr.Image()
        with gr.Column():
            question_input = gr.Textbox(label="Question")
            und_seed_input = gr.Number(label="Seed", precision=0, value=42)
            top_p = gr.Slider(minimum=0, maximum=1, value=0.95, step=0.05, label="top_p")
            temperature = gr.Slider(minimum=0, maximum=1, value=0.1, step=0.05, label="temperature")
        
    understanding_button = gr.Button("Chat")
    understanding_output = gr.Textbox(label="Response")

    examples_inpainting = gr.Examples(
        label="Multimodal Understanding examples",
        examples=[
            [
                "explain this meme",
                "./images/doge.png",
            ],
            [
                "Convert the formula into latex code.",
                "./images/equation.png",
            ],
        ],
        inputs=[question_input, image_input],
    )
    
        
    gr.Markdown(value="# Text-to-Image Generation")

    
    
    with gr.Row():
        cfg_weight_input = gr.Slider(minimum=1, maximum=10, value=2, step=0.5, label="CFG Weight")
        step_input = gr.Slider(minimum=1, maximum=50, value=30, step=1, label="Number of Inference Steps")

    prompt_input = gr.Textbox(label="Prompt")
    seed_input = gr.Number(label="Seed (Optional)", precision=0, value=12345)

    generation_button = gr.Button("Generate Images")

    image_output = gr.Gallery(label="Generated Images", columns=2, rows=2, height=300)

    examples_t2i = gr.Examples(
        label="Text to image generation examples.",
        examples=[
            "Master shifu racoon wearing drip attire as a street gangster.",
            "A cute and adorable baby fox with big brown eyes, autumn leaves in the background enchanting,immortal,fluffy, shiny mane,Petals,fairyism,unreal engine 5 and Octane Render,highly detailed, photorealistic, cinematic, natural colors.",
            "The image features an intricately designed eye set against a circular backdrop adorned with ornate swirl patterns that evoke both realism and surrealism. At the center of attention is a strikingly vivid blue iris surrounded by delicate veins radiating outward from the pupil to create depth and intensity. The eyelashes are long and dark, casting subtle shadows on the skin around them which appears smooth yet slightly textured as if aged or weathered over time.\n\nAbove the eye, there's a stone-like structure resembling part of classical architecture, adding layers of mystery and timeless elegance to the composition. This architectural element contrasts sharply but harmoniously with the organic curves surrounding it. Below the eye lies another decorative motif reminiscent of baroque artistry, further enhancing the overall sense of eternity encapsulated within each meticulously crafted detail. \n\nOverall, the atmosphere exudes a mysterious aura intertwined seamlessly with elements suggesting timelessness, achieved through the juxtaposition of realistic textures and surreal artistic flourishes. Each component\u2014from the intricate designs framing the eye to the ancient-looking stone piece above\u2014contributes uniquely towards creating a visually captivating tableau imbued with enigmatic allure.",
        ],
        inputs=prompt_input,
    )
    
    understanding_button.click(
        multimodal_understanding,
        inputs=[image_input, question_input, und_seed_input, top_p, temperature],
        outputs=understanding_output
    )
    
    generation_button.click(
        fn=generate_image,
        inputs=[prompt_input, seed_input, cfg_weight_input, step_input],
        outputs=image_output
    )

demo.launch(share=True)


# Copyright 2025 ByteDance and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import os
import subprocess

import numpy as np
from scipy.io import wavfile
import pyloudnorm as pyln
from pydub import AudioSegment


def to_wav_bytes(wav, sr, norm=False):
    wav = wav.astype(float)
    if norm:
        meter = pyln.Meter(sr)  # create BS.1770 meter
        loudness = meter.integrated_loudness(wav)
        wav = pyln.normalize.loudness(wav, loudness, -18.0)
        if np.abs(wav).max() >= 1:
            wav = wav / np.abs(wav).max() * 0.95
    wav = wav * 32767
    bytes_io = io.BytesIO()
    wavfile.write(bytes_io, sr, wav.astype(np.int16))
    return bytes_io.getvalue()


def save_wav(wav_bytes, path):
    with open(path[:-4] + '.wav', 'wb') as file:
        file.write(wav_bytes)
    if path[-4:] == '.mp3':
        to_mp3(path[:-4])


def to_mp3(out_path):
    if out_path[-4:] == '.wav':
        out_path = out_path[:-4]
    subprocess.check_call(
        f'ffmpeg -threads 1 -loglevel error -i "{out_path}.wav" -vn -b:a 192k -y -hide_banner -async 1 "{out_path}.mp3"',
        shell=True, stdin=subprocess.PIPE)
    subprocess.check_call(f'rm -f "{out_path}.wav"', shell=True)


def convert_to_wav(wav_path):
    # Check if the file exists
    if not os.path.exists(wav_path):
        print(f"The file '{wav_path}' does not exist.")
        return

    # Check if the file already has a .wav extension
    if not wav_path.endswith(".wav"):
        # Define the output path with a .wav extension
        out_path = os.path.splitext(wav_path)[0] + ".wav"

        # Load the audio file using pydub and convert it to WAV
        audio = AudioSegment.from_file(wav_path)
        audio.export(out_path, format="wav")

        print(f"Converted '{wav_path}' to '{out_path}'")


def convert_to_wav_bytes(audio_binary):
    # Load the audio binary using pydub and convert it to WAV
    audio = AudioSegment.from_file(io.BytesIO(audio_binary))
    wav_bytes = io.BytesIO()
    audio.export(wav_bytes, format="wav")
    wav_bytes.seek(0)
    return wav_bytes


''' Smoothly combine audio segments using crossfade transitions." '''
def combine_audio_segments(segments, crossfade_duration=0.16, sr=24000):
    window_length = int(sr * crossfade_duration)
    hanning_window = np.hanning(2 * window_length)
    # Combine
    for i, segment in enumerate(segments):
        if i == 0:
            combined_audio = segment
        else:
            overlap = combined_audio[-window_length:] * hanning_window[window_length:] + segment[:window_length] * hanning_window[:window_length]
            combined_audio = np.concatenate(
                [combined_audio[:-window_length], overlap, segment[window_length:]]
            )
    return combined_audio


import logging
from typing import AsyncGenerator

import aiohttp
from apipeline.frames.sys_frames import ErrorFrame
from apipeline.frames.data_frames import Frame, AudioRawFrame

from achatbot.processors.speech.tts.base import TTSProcessorBase
from types.frames.control_frames import TTSStartedFrame


class DeepgramTTSProcessor(TTSProcessorBase):
    def __init__(
        self,
        *,
        api_key: str,
        voice: str = "aura-helios-en",
        base_url: str = "https://api.deepgram.com/v1/speak",
        sample_rate: int = 16000,
        encoding: str = "linear16",
        aiohttp_session: aiohttp.ClientSession | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self._voice = voice
        self._api_key = api_key
        self._base_url = base_url
        self._sample_rate = sample_rate
        self._encoding = encoding
        self._aiohttp_session = aiohttp_session or aiohttp.ClientSession()
        self._close_aiohttp_session = aiohttp_session is None

    def can_generate_metrics(self) -> bool:
        return True

    async def cleanup(self):
        await super().cleanup()
        if self._close_aiohttp_session:
            await self._aiohttp_session.close()

    async def set_voice(self, voice: str):
        logging.debug(f"Switching TTS voice to: [{voice}]")
        self._voice = voice

    async def run_tts(self, text: str) -> AsyncGenerator[Frame, None]:
        logging.debug(f"Generating TTS: [{text}]")

        base_url = self._base_url
        request_url = f"{base_url}?model={self._voice}&encoding={self._encoding}&container=none&sample_rate={self._sample_rate}"
        headers = {"authorization": f"token {self._api_key}"}
        body = {"text": text}

        try:
            await self.start_ttfb_metrics()
            async with self._aiohttp_session.post(request_url, headers=headers, json=body) as r:
                if r.status != 200:
                    response_text = await r.text()
                    # If we get a a "Bad Request: Input is unutterable", just print out a debug log.
                    # All other unsuccesful requests should emit an error frame. If not specifically
                    # handled by the running PipelineTask, the ErrorFrame will cancel the task.
                    if "unutterable" in response_text:
                        logging.debug(f"Unutterable text: [{text}]")
                        return

                    logging.error(
                        f"{self} error getting audio (status: {r.status}, error: {response_text})"
                    )
                    yield ErrorFrame(
                        f"Error getting audio (status: {r.status}, error: {response_text})"
                    )
                    return

                await self.start_tts_usage_metrics(text)

                async for data in r.content:
                    await self.stop_ttfb_metrics()
                    frame = AudioRawFrame(audio=data, sample_rate=self._sample_rate, num_channels=1)
                    yield frame
        except Exception as e:
            logging.exception(f"{self} exception: {e}")
        finally:
            self._tts_done_event.set()



import base64
import json
import os
import logging
from typing import List


try:
    from openai import AsyncStream
    import litellm
    from openai.types.chat import (
        ChatCompletionChunk,
        ChatCompletionFunctionMessageParam,
        ChatCompletionMessageParam,
        ChatCompletionToolParam,
    )
except ModuleNotFoundError as e:
    logging.error(f"Exception: {e}")
    logging.error(
        "In order to use OpenAI, you need to `pip install achatbot[litellm_processor]`. "
        "Also, set environment variable such as `OPENAI_API_KEY`."
        "see: https://docs.litellm.ai/docs/"
    )
    raise Exception(f"Missing module: {e}")
import httpx
from apipeline.frames.data_frames import TextFrame, Frame
from apipeline.pipeline.pipeline import FrameDirection

from src.processors.aggregators.openai_llm_context import OpenAILLMContext, OpenAILLMContextFrame
from src.processors.llm.base import LLMProcessor, UnhandledFunctionException
from src.types.frames.control_frames import (
    LLMFullResponseEndFrame,
    LLMFullResponseStartFrame,
    LLMModelUpdateFrame,
)
from src.types.frames.data_frames import LLMMessagesFrame, VisionImageRawFrame


class LiteLLMProcessor(LLMProcessor):
    """
    llm proxy to use the OpenAI Input/Output Format
    see: https://docs.litellm.ai/docs/
    """

    TAG = "litellm_processor"

    def __init__(self, model: str = "github/gpt-4o", set_verbose: bool = False, **kwargs):
        super().__init__(model=model, **kwargs)
        super().__init__(**kwargs)
        self._model: str = model
        litellm.set_verbose = set_verbose
        # self._response_format = {"type": "json_object"}

    async def call_function(
        self, *, context: OpenAILLMContext, tool_call_id: str, function_name: str, arguments: dict
    ) -> None:
        f = None
        if function_name in self._callbacks.keys():
            f = self._callbacks[function_name]
        elif None in self._callbacks.keys():
            f = self._callbacks[None]
        else:
            return None
        await context.call_function(
            f, function_name=function_name, tool_call_id=tool_call_id, arguments=arguments, llm=self
        )

    # QUESTION FOR CB: maybe this isn't needed anymore?
    async def call_start_function(self, context: OpenAILLMContext, function_name: str):
        if function_name in self._start_callbacks.keys():
            await self._start_callbacks[function_name](function_name, self, context)
        elif None in self._start_callbacks.keys():
            return await self._start_callbacks[None](function_name, self, context)

    async def get_chat_completions(
        self, context: OpenAILLMContext, messages: List[ChatCompletionMessageParam]
    ) -> AsyncStream[ChatCompletionChunk]:
        chunks = await litellm.acompletion(
            model=self._model,
            stream=True,
            # trim_messages ensures tokens(messages) < max_tokens(model)
            # need HF_TOKEN to be set
            # issue: https://github.com/BerriAI/litellm/issues/6505
            # messages=litellm.utils.trim_messages(messages, self._model),
            messages=messages,
            tools=context.tools,
            tool_choice=context.tool_choice,
            stream_options={"include_usage": True},
            # stream Structured Outputs json mode !TODO @weedge -> partial support
            # https://docs.litellm.ai/docs/completion/json_mode
            # response_format={"type": "json_object"}  # 👈 KEY CHANGE
        )
        return chunks

    async def _stream_chat_completions(
        self, context: OpenAILLMContext
    ) -> AsyncStream[ChatCompletionChunk]:
        logging.info(f"Generating chat context messages: {context.get_messages_json()}")

        messages: List[ChatCompletionMessageParam] = context.get_messages()

        # base64 encode any images
        for message in messages:
            if message.get("mime_type") == "image/jpeg":
                encoded_image = base64.b64encode(message["data"].getvalue()).decode("utf-8")
                text = message["content"]
                message["content"] = [
                    {"type": "text", "text": text},
                    {
                        "type": "image_url",
                        "image_url": {"url": f"data:image/jpeg;base64,{encoded_image}"},
                    },
                ]
                del message["data"]
                del message["mime_type"]

        chunks = await self.get_chat_completions(context, messages)

        return chunks

    async def record_llm_usage_tokens(self, chunk_dict: dict):
        tokens = {
            "processor": self.name,
            "model": self._model,
        }
        if hasattr(chunk_dict, "usage") and chunk_dict["usage"]:
            tokens["prompt_tokens"] = chunk_dict["usage"]["prompt_tokens"]
            tokens["completion_tokens"] = chunk_dict["usage"]["completion_tokens"]
            tokens["total_tokens"] = chunk_dict["usage"]["total_tokens"]
            await self.start_llm_usage_metrics(tokens)

    async def _process_context(self, context: OpenAILLMContext):
        function_name = ""
        arguments = ""
        tool_call_id = ""

        await self.start_ttfb_metrics()

        chunk_stream: AsyncStream[ChatCompletionChunk] = await self._stream_chat_completions(
            context
        )

        async for chunk in chunk_stream:
            # logging.info(f"chunk:{chunk.model_dump_json()}")
            await self.record_llm_usage_tokens(chunk_dict=chunk.model_dump())

            if len(chunk.choices) == 0:
                continue
            await self.stop_ttfb_metrics()

            if chunk.choices[0].delta.tool_calls:
                # We're streaming the LLM response to enable the fastest response times.
                # For text, we just yield each chunk as we receive it and count on consumers
                # to do whatever coalescing they need (eg. to pass full sentences to TTS)
                #
                # If the LLM is a function call, we'll do some coalescing here.
                # If the response contains a function name, we'll yield a frame to tell consumers
                # that they can start preparing to call the function with that name.
                # We accumulate all the arguments for the rest of the streamed response, then when
                # the response is done, we package up all the arguments and the function name and
                # yield a frame containing the function name and the arguments.

                tool_call = chunk.choices[0].delta.tool_calls[0]
                if tool_call.function and tool_call.function.name:
                    function_name += tool_call.function.name
                    tool_call_id = tool_call.id
                    await self.call_start_function(context, function_name)
                if tool_call.function and tool_call.function.arguments:
                    # Keep iterating through the response to collect all the argument fragments
                    arguments += tool_call.function.arguments
            elif chunk.choices[0].delta.content:
                await self.push_frame(TextFrame(chunk.choices[0].delta.content))

        # if we got a function name and arguments, check to see if it's a function with
        # a registered handler. If so, run the registered callback, save the result to
        # the context, and re-prompt to get a chat answer. If we don't have a registered
        # handler, raise an exception.
        if function_name and arguments:
            if self.has_function(function_name):
                await self._handle_function_call(context, tool_call_id, function_name, arguments)
            else:
                raise UnhandledFunctionException(
                    f"The LLM tried to call a function named '{function_name}', but there isn't a callback registered for that function."
                )

    async def _handle_function_call(
        self, context: OpenAILLMContext, tool_call_id: str, function_name: str, arguments: str
    ):
        arguments = json.loads(arguments)
        await self.call_function(
            context=context,
            tool_call_id=tool_call_id,
            function_name=function_name,
            arguments=arguments,
        )

    async def process_frame(self, frame: Frame, direction: FrameDirection):
        await super().process_frame(frame, direction)

        context = None
        if isinstance(frame, OpenAILLMContextFrame):
            context: OpenAILLMContext = frame.context
        elif isinstance(frame, LLMMessagesFrame):
            context = OpenAILLMContext.from_messages(frame.messages)
        elif isinstance(frame, VisionImageRawFrame):
            context = OpenAILLMContext.from_image_frame(frame)
        elif isinstance(frame, LLMModelUpdateFrame):
            logging.debug(f"Switching LLM model to: [{frame.model}]")
            self._model = frame.model
        else:
            await self.push_frame(frame, direction)

        if context:
            await self.push_frame(LLMFullResponseStartFrame())
            await self.start_processing_metrics()
            await self._process_context(context)
            await self.stop_processing_metrics()
            await self.push_frame(LLMFullResponseEndFrame())



# Copyright (c) 2021 Mobvoi Inc. (authors: Binbin Zhang)
#               2024 Alibaba Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import torchaudio
import logging
logging.getLogger('matplotlib').setLevel(logging.WARNING)
logging.basicConfig(level=logging.DEBUG,
                    format='%(asctime)s %(levelname)s %(message)s')

def read_trans(list_file):
    trans = {}
    with open(list_file, 'r', encoding='utf8') as fin:
        for line in fin:
            sec = line.strip().split("\t")
            if len(sec) > 1:
                if sec[0] not in trans.keys():
                    trans[sec[0]] = sec[1]
    return trans

def read_scp(list_file):
    scp = {}
    with open(list_file, 'r', encoding='utf8') as fin:
        for line in fin:
            sec = line.strip().split(" ")
            if len(sec) > 1:
                if sec[0] not in scp.keys():
                    scp[sec[0]] = sec[1]
    return scp

def read_lists(list_file):
    lists = []
    with open(list_file, 'r', encoding='utf8') as fin:
        for line in fin:
            lists.append(line.strip())
    return lists


def read_json_lists(list_file):
    lists = read_lists(list_file)
    results = {}
    for fn in lists:
        with open(fn, 'r', encoding='utf8') as fin:
            results.update(json.load(fin))
    return results


def load_wav(wav, target_sr):
    audio, sample_rate = torchaudio.load(wav)
    audio = audio.mean(dim=0, keepdim=True)
    if sample_rate != target_sr:
        assert sample_rate > target_sr, 'wav sample rate {} must be greater than {}'.format(sample_rate, target_sr)
        audio = torchaudio.transforms.Resample(orig_freq=sample_rate, new_freq=target_sr)(speech)
    return audio


def speed_change(waveform, sample_rate, speed_factor: str):
    effects = [
        ["tempo", speed_factor],  # speed_factor
        ["rate", f"{sample_rate}"]
    ]
    augmented_waveform, new_sample_rate = torchaudio.sox_effects.apply_effects_tensor(
        waveform,
        sample_rate,
        effects
    )
    return augmented_waveform, new_sample_rate



# Copyright 2025 ByteDance and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch

def mel2token_to_dur(mel2token, T_txt=None, max_dur=None):
    is_torch = isinstance(mel2token, torch.Tensor)
    has_batch_dim = True
    if not is_torch:
        mel2token = torch.LongTensor(mel2token)
    if T_txt is None:
        T_txt = mel2token.max()
    if len(mel2token.shape) == 1:
        mel2token = mel2token[None, ...]
        has_batch_dim = False
    B, _ = mel2token.shape
    dur = mel2token.new_zeros(B, T_txt + 1).scatter_add(1, mel2token, torch.ones_like(mel2token))
    dur = dur[:, 1:]
    if max_dur is not None:
        dur = dur.clamp(max=max_dur)
    if not is_torch:
        dur = dur.numpy()
    if not has_batch_dim:
        dur = dur[0]
    return dur



from argparse import ArgumentParser
from http import HTTPStatus
from typing import Annotated, Any

import ormsgpack
from baize.datastructures import ContentType
from kui.asgi import HTTPException, HttpRequest

from fish_speech.inference_engine import TTSInferenceEngine
from fish_speech.utils.schema import ServeTTSRequest
from tools.server.inference import inference_wrapper as inference


def parse_args():
    parser = ArgumentParser()
    parser.add_argument("--mode", type=str, choices=["agent", "tts"], default="tts")
    parser.add_argument("--load-asr-model", action="store_true")
    parser.add_argument(
        "--llama-checkpoint-path",
        type=str,
        default="checkpoints/fish-speech-1.5",
    )
    parser.add_argument(
        "--decoder-checkpoint-path",
        type=str,
        default="checkpoints/fish-speech-1.5/firefly-gan-vq-fsq-8x1024-21hz-generator.pth",
    )
    parser.add_argument("--decoder-config-name", type=str, default="firefly_gan_vq")
    parser.add_argument("--device", type=str, default="cuda")
    parser.add_argument("--half", action="store_true")
    parser.add_argument("--compile", action="store_true")
    parser.add_argument("--max-text-length", type=int, default=0)
    parser.add_argument("--listen", type=str, default="127.0.0.1:8080")
    parser.add_argument("--workers", type=int, default=1)
    parser.add_argument("--api-key", type=str, default=None)

    return parser.parse_args()


class MsgPackRequest(HttpRequest):
    async def data(
        self,
    ) -> Annotated[
        Any, ContentType("application/msgpack"), ContentType("application/json")
    ]:
        if self.content_type == "application/msgpack":
            return ormsgpack.unpackb(await self.body)

        elif self.content_type == "application/json":
            return await self.json

        raise HTTPException(
            HTTPStatus.UNSUPPORTED_MEDIA_TYPE,
            headers={"Accept": "application/msgpack, application/json"},
        )


async def inference_async(req: ServeTTSRequest, engine: TTSInferenceEngine):
    for chunk in inference(req, engine):
        if isinstance(chunk, bytes):
            yield chunk


async def buffer_to_async_generator(buffer):
    yield buffer


def get_content_type(audio_format):
    if audio_format == "wav":
        return "audio/wav"
    elif audio_format == "flac":
        return "audio/flac"
    elif audio_format == "mp3":
        return "audio/mpeg"
    else:
        return "application/octet-stream"



from __future__ import print_function

import builtins
import datetime
import time
import math
import argparse
import threading
from copy import deepcopy
from dataclasses import dataclass
from typing import Generator

import soundfile as sf
import numpy as np
import torch
import torchaudio
import torchaudio.compliance.kaldi as k

from bin.inference import audioEncoderProcessor
from models.utils import print_outputs
from web.queue import PCMQueue, ThreadSafeQueue
from models.pipeline import inferencePipeline
from models.decoder.llm2tts import llm2TTS


def custom_print(*args, **kwargs):
    current_time = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")[:-3]
    original_print(f"[{current_time}]", *args, **kwargs)


def get_args():
    parser = argparse.ArgumentParser(description="Freeze-Omni-Inference-stream")
    parser.add_argument("--model_path", required=True, help="model_path to load")
    parser.add_argument("--llm_path", required=True, help="llm_path to load")
    parser.add_argument("--top_k", type=int, default=5)
    parser.add_argument("--top_p", type=float, default=0.8)
    parser.add_argument("--temperature", type=float, default=0.7)
    parser.add_argument("--input_wav", required=True, help="input wav")
    parser.add_argument("--output_wav", required=True, help="output wav")

    args = parser.parse_args()
    print(args)
    return args


class GlobalVars:
    """
    multi turn conversation interal speech dialogue outputs for listen and speak
    """

    speech_dialogue_outputs = {}

    @staticmethod
    def deepcopy_outputs():  # cp for write
        return deepcopy(GlobalVars.speech_dialogue_outputs)


@dataclass
class PCMStatChunk:
    status: str  # sl(start listen), cl(continue listen), el(end listen)
    data: torch.Tensor  # chunk tensor

    def __post_init__(self):
        if self.status not in ["sl", "cl", "el"]:
            raise ValueError("status must be one of: 'sl','cl','el'")

    def __str__(self):
        return (
            f"status:{self.status} data shape:{self.data.shape if self.data is not None else None}"
        )


@dataclass
class Listen2SpeakFrame:
    """
    outputs have stat, so need pass it from listen to speak
    """

    outputs: dict

    def __str__(self):
        return f"outputs:{print_outputs(self.outputs)}"


@dataclass
class GenTTSFrame:
    text: str
    data: np.ndarray  # a numpy array of dtype np.float32

    def __str__(self):
        return f"text:{self.text} numpy data:{self.data.shape if self.data is not None else None}"


class PCMListener:
    def __init__(
        self,
        outputs_queue: ThreadSafeQueue,  # outputs -> out queue
        pipeline: inferencePipeline,
        cache_history_size: int = 10,
        system_prompt: str = "You are a helpful assistant.",
    ) -> None:
        self.pcm_stat_chunk_queue = ThreadSafeQueue()  # pcm data -> in queue

        self.outputs_queue = outputs_queue

        # stream chunk to encoder
        self.audio_processor = audioEncoderProcessor()
        # encoder and audio llm
        self.pipeline = pipeline

        # pre status system prompt, outputs stat: pre -> sl
        self.outputs = self.pipeline.speech_dialogue(None, stat="pre", role=system_prompt)
        # print(f"pre-> outputs:[{print_outputs(self.outputs)}]")
        GlobalVars.speech_dialogue_outputs = self.outputs

        # chunck feat history cache also use ring buffer to do :)
        self.history = torch.zeros(
            [
                cache_history_size,
                self.audio_processor.chunk_size + self.audio_processor.chunk_overlap,
                self.audio_processor.feat_dim,
            ]
        )

        # start listen thread
        self.listen_thread = threading.Thread(target=self.listen, args=())
        self.stop_listen = False
        self.listen_thread.start()

    def stop(self):
        self.stop_listen = True
        self.listen_thread.join(timeout=3)

    def send(self, pcm_items: torch.Tensor | None, status: str):
        if pcm_items is None:
            item = PCMStatChunk(status=status, data=None)
            self.pcm_stat_chunk_queue.put(item)
        else:
            item = PCMStatChunk(status=status, data=pcm_items)
            self.pcm_stat_chunk_queue.put(item)

    def history_buffering_strategy(self, input_chunk: torch.Tensor) -> torch.Tensor:
        # cache fbank feature (input_chunk)
        # << 1
        self.history[:-1] = self.history[1:].clone()
        # last history = input chunk
        self.history[-1:] = input_chunk
        # start listen to # copy last 6 chunks
        feature_last_chunk = self.history[-6:].unsqueeze(1)
        return feature_last_chunk

    def llm_prefill(self, status: str, feature: torch.Tensor, outputs: dict, is_first_pack=False):
        """
        Prefills the LLM of speech dialogue system using speech.

        Parameters:
        - status: the current state of the audio input.
        - feature: audio feature tensor.
        - outputs (dict): A dictionary containing the current state of the dialogue system.
        - is_first_pack (bool, optional): Indicates if the current input packet is the first one in a new conversation
        """

        if status == "sl":
            # Satge1: start listen
            # stat will be auto set to 'cl' after Stage1
            outputs = self.pipeline.speech_dialogue(feature, **outputs)
            return outputs
        if status == "el":
            print("status end listen. start to speak")
            return outputs

        if status == "cl":
            if outputs["stat"] == "cl" or outputs["stat"] == "sl":
                # Stage2: continue listen
                # stat will be auto set to 'ss' when endpoint is detected
                # print("output stat continue listen")
                outputs = self.pipeline.speech_dialogue(feature, **outputs)
            if is_first_pack:
                outputs["stat"] = "cl"
            if outputs["stat"] == "el":
                print("output stat end listen. Detect invalid break")
            if outputs["stat"] == "ss":
                # print(f"start speak. start to speak")
                pass
        return outputs

    def listen(self):
        """
        chunk status from VAD
        """
        print("Start listening")
        while True:
            if self.stop_listen:
                print("Stop listening")
                break
            stat_chunk: PCMStatChunk = self.pcm_stat_chunk_queue.get()
            if stat_chunk is None:
                time.sleep(0.01)
                continue
            # print(f"Received PCM stat chunk: {stat_chunk}")

            # if self.outputs['stat'] =="sl":
            #    print(f"stat_chunk data:{stat_chunk.data}")

            if stat_chunk.status == "sl":
                fbank_feature = self.audio_processor.process(stat_chunk.data)
                self.outputs = (
                    GlobalVars.deepcopy_outputs()
                )  # for next turn conversation outputs from speak where in gloab var
                # print(f"sl-> outputs:{print_outputs(self.outputs)}")
                self.outputs["adapter_cache"] = None
                self.outputs["encoder_cache"] = None
                self.outputs["pe_index"] = 0
                self.outputs["stat"] = "sl"
                self.outputs["last_id"] = None
                if "text" in self.outputs:
                    del self.outputs["text"]
                if "hidden_state" in self.outputs:
                    del self.outputs["hidden_state"]

                self.outputs = self.llm_prefill("sl", fbank_feature, self.outputs)

            elif stat_chunk.status == "cl":
                # print(f"cl in-> outputs:{print_outputs(self.outputs)}")
                fbank_feature = self.audio_processor.process(stat_chunk.data)
                # if self.outputs['stat'] =="sl":
                #    print(f"fbank:{fbank_feature}")
                self.outputs = self.llm_prefill(stat_chunk.status, fbank_feature, self.outputs)
                # print(f"cl out-> outputs:{print_outputs(self.outputs)}")
                if self.outputs["stat"] == "ss":
                    pass
                else:
                    self.outputs["stat"] = "cl"
            elif stat_chunk.status == "el":
                self.audio_processor.reset()

                # print(f"el-> outputs:[{print_outputs(self.outputs)}]")
                self.outputs["adapter_cache"] = None
                self.outputs["encoder_cache"] = None
                self.outputs["pe_index"] = 0
                self.outputs["last_id"] = None
                self.outputs["stat"] = "ss"

            if self.outputs["stat"] == "ss":
                frame = Listen2SpeakFrame(self.outputs)
                # print(f"start to speak, send frame:[{frame}]")
                self.outputs_queue.put(frame)


@dataclass
class TTSSpeakerArgs:
    # https://huggingface.co/VITA-MLLM/Freeze-Omni/blob/main/checkpoints/server.json
    # decoder(LLM2TTSCodecAR)
    # NAR llama transformer decoder pre_nn_forward -> NAR llama transformer decoder kv_cache_prefix_forward -> AR llama transformer decoder transformer_infer

    # llama transformer decoder
    decoder_top_k: int = 2
    decoder_penalty_window_size: int = -1  # <0 no penalty window
    decoder_penalty: float = 1.1

    # codec decoder
    decoder_first_chunk_size: int = 20
    decoder_chunk_size: int = 40
    decoder_chunk_overlap_size: int = 10

    # find_min_sum_index
    decoder_N: int = 2401
    decoder_seg_threshold_first_pack: float = 0.1
    decoder_seg_threshold: float = 0.015


class TTSSpeaker:
    def __init__(
        self,
        args: TTSSpeakerArgs,
        outputs_in_queue: ThreadSafeQueue,  # out queue -> speak
        pipeline: inferencePipeline,
        tts: llm2TTS,
    ) -> None:
        self.args = args
        self.outputs_in_queue = outputs_in_queue
        self.pipeline = pipeline
        self.tts = tts

        self.reset()

        # speak thread
        self.speak_thread = threading.Thread(target=self.speak, args=())
        self.speak_thread.start()

    def reset(self):
        self.stop_speak = False
        self.is_generate = False
        self.whole_text = ""

        self.tts_over = False
        self.tts_over_time = 0
        self.tts_data = ThreadSafeQueue()

        self.stop_tts = False

    def print(self):
        print("stop_speak:", self.stop_speak)
        print("is_generate:", self.is_generate)
        print("whole_text:", self.whole_text)
        print("tts_over:", self.tts_over)
        print("tts_over_time:", self.tts_over_time)

    @property
    def gen_text(self):
        """Get the whole text."""
        return self.whole_text

    def stop(self):
        """Stop the speak thread."""
        self.stop_speak = True
        self.speak_thread.join()

    def decoder(
        self, cur_hidden_state: list[torch.Tensor], cur_text: str, generate_num: int
    ) -> int:
        """
        Decodes the current hidden state and text to generate audio segments using speech decoder.

        Parameters:
        - cur_hidden_state (list of torch.Tensor): The current hidden state of the language model.
        - cur_text (str): The current text to be synthesized.
        - generate_num (int): The number of audio segments generated

        Returns:
        - int: The updated number of audio segments generated.
        """
        hidden_state_output = torch.cat(cur_hidden_state).squeeze(1)
        cur_text_procced = self.pipeline.post_process(cur_text)
        print("Synthesis: ", [cur_text_procced])
        embeddings = self.pipeline.model.llm_decoder.model.embed_tokens(
            torch.tensor(self.pipeline.model.tokenizer.encode(cur_text_procced)).cuda()
        )
        codec_chunk_size = self.args.decoder_first_chunk_size
        codec_padding_size = self.args.decoder_chunk_overlap_size
        seg_threshold = self.args.decoder_seg_threshold_first_pack
        if generate_num != 0:
            codec_chunk_size = self.args.decoder_chunk_size
            seg_threshold = self.args.decoder_seg_threshold
        for seg in self.tts.run(
            embeddings.reshape(-1, 896).unsqueeze(0),
            self.args.decoder_top_k,
            hidden_state_output.reshape(-1, 896).unsqueeze(0),
            codec_chunk_size=codec_chunk_size,
            codec_padding_size=codec_padding_size,
            penalty_window_size=self.args.decoder_penalty_window_size,
            penalty=self.args.decoder_penalty,
            N=self.args.decoder_N,
            seg_threshold=seg_threshold,
        ):
            if generate_num == 0:
                try:
                    split_idx = torch.nonzero(seg.abs() > 0.03, as_tuple=True)[-1][0]
                    seg = seg[:, :, split_idx:]
                except Exception:
                    print("Do not need to split")
                    pass
            generate_num += 1
            if self.tts_over:
                self.tts_data.clear()
                self.whole_text = ""
                self.tts_data.put(GenTTSFrame(text="", data=None))
                break
            frame = GenTTSFrame(text=cur_text, data=seg.squeeze().float().cpu().numpy())
            self.tts_data.put(frame)
        return generate_num

    def get_tts_data(self) -> Generator[GenTTSFrame, None, None]:
        """
        get tts bytes data
        """
        while True:
            if self.stop_speak:
                print("Stop speak so break get tts data")
                break
            output_data = self.tts_data.get()
            if output_data is not None:
                # print("Get TTS data")
                # yield output_data.astype(np.int16).tobytes()
                yield output_data
            else:
                yield None

    def speak(self):
        """
        Generates speech dialogue output based on the current state
        """
        while True:
            if self.stop_speak:
                print("Stop speak")
                break
            frame: Listen2SpeakFrame = self.outputs_in_queue.get()
            if frame is None:
                time.sleep(0.01)
                continue

            outputs = frame.outputs
            # print(f"speak get-> outputs:[{print_outputs(outputs)}]")
            # Stage3: start speak
            self.is_generate = True
            outputs = self.pipeline.speech_dialogue(None, **outputs)
            # print(f"ss-> outputs:[{print_outputs(outputs)}]")

            cur_hidden_state = []
            cur_hidden_state.append(outputs["hidden_state"])

            # Stage4: contiune speak until stat is set to 'sl'
            # use 'stop' to interrupt generation, stat need to be manually set as 'sl'
            stop = False
            cur_text = ""
            last_text = ""
            generate_num = 0
            while True:
                if self.stop_speak:
                    break
                if len(outputs["past_tokens"]) > 500:
                    stop = True
                if stop:
                    break
                del outputs["text"]
                del outputs["hidden_state"]
                outputs = self.pipeline.speech_dialogue(None, **outputs)
                # print(f"sc-> outputs:[{print_outputs(outputs)}]")
                if outputs["stat"] == "cs":
                    cur_hidden_state.append(outputs["hidden_state"])
                    if "�" in outputs["text"][len(last_text) :]:
                        continue
                    self.whole_text += outputs["text"][len(last_text) :]
                    cur_text += outputs["text"][len(last_text) :]
                    # print(self.whole_text])
                    if generate_num == 0 or (len(cur_hidden_state) >= 20):
                        suffix_list = [",", "，", "。", "：", "？", "！", ".", ":", "?", "!", "\n"]
                    else:
                        suffix_list = ["。", "：", "？", "！", ".", "?", "!", "\n"]
                    if outputs["text"][len(last_text) :].endswith(tuple(suffix_list)) and (
                        len(cur_hidden_state) >= 4
                    ):
                        if (
                            outputs["text"][len(last_text) :].endswith(".")
                            and last_text[-1].isdigit()
                        ):
                            pass
                        else:
                            if not self.tts_over:
                                if len(cur_hidden_state) > 0:
                                    generate_num = self.decoder(
                                        cur_hidden_state, cur_text, generate_num
                                    )
                                    cur_text = ""
                                    cur_hidden_state = []
                    last_text = outputs["text"]
                else:
                    break
            if not self.tts_over:
                if len(cur_hidden_state) != 0:
                    generate_num = self.decoder(cur_hidden_state, cur_text, generate_num)
                    cur_text = ""
            self.is_generate = False
            outputs["stat"] = "sl"
            outputs["last_id"] = None
            GlobalVars.speech_dialogue_outputs = deepcopy(outputs)
            print(self.whole_text)
            self.tts_data.put(GenTTSFrame(text="", data=None))

    def interrupt(self):
        self.stop_speak = True
        self.tts_over = True
        while True:
            time.sleep(0.01)
            if self.is_generate is False:
                self.stop_speak = False
                while True:
                    time.sleep(0.01)
                    if self.tts_data.is_empty():
                        self.whole_text = ""
                        self.tts_over = False
                        self.tts_over_time += 1
                        break
                break


def inference_stream(listener: PCMListener, speaker: TTSSpeaker, configs):
    """
    Perform inference for a speech dialogue system.
    - 流式语音输入通过语音编码器形成chunk-wise特征，然后通过适配器连接到LLM。
    - LLM生成隐藏状态和文本标记，在块分割后分别以块的形式发送到NAR前缀语音解码器和NAR语音解码器。
    - 最后，AR语音解码器将生成的令牌发送到语音令牌FIFO中，流式编解码器根据固定的语音令牌块大小从FIFO生成流式语音输出。

    Parameters:
    - listener: listen pcm data(chunk) to asr with pipeline(encoder and adpter),
        - detail status: sl, cl, el, ss
    - speaker: tts speaker with pipeline(audio llm) and decoder (NAR decoder AR decoder and codec decoder)
        - detail status: ss, cs
    - configs: Input args. (argparse)

    Returns:
    - None
    """
    wav, fs = sf.read(configs.input_wav)
    wav = torch.tensor(wav)
    if fs != 16000:
        wav = torchaudio.transforms.Resample(orig_freq=fs, new_freq=16000)(wav.float())
        fs = 16000

    # like io_uring
    chunk_size = listener.audio_processor.get_chunk_size()
    wav_input = torch.zeros(math.ceil(wav.shape[0] / chunk_size) * chunk_size)
    wav_input[: wav.shape[0]] = wav
    for i in range(0, wav_input.shape[0], chunk_size):
        print("--->", wav_input.shape, wav.shape, wav_input[i : i + chunk_size].shape)
        status = "cl"
        if i == 0:
            status = "sl"
        # send pcm data with status
        listener.send(wav_input[i : i + chunk_size], status)
    listener.send(None, "el")

    wavs = []
    # get tts speak data
    for item in speaker.get_tts_data():
        if item:
            print(item)
            if item.data is None:
                break
            else:
                wavs.append(item.data)
        else:
            time.sleep(0.01)  # yield thread

    sf.write(configs.output_wav, np.concatenate(wavs, -1), 24000)
    print(f"write to {configs.output_wav}")

    listener.stop()
    speaker.stop()


if __name__ == "__main__":
    # change print function to add time stamp
    original_print = builtins.print
    builtins.print = custom_print

    configs = get_args()

    # encoder and audio llm
    pipeline = inferencePipeline(configs)
    # decoder
    tts = llm2TTS(configs.model_path)
    # listen -> gen_queue -> speak
    gen_queue = ThreadSafeQueue()
    # listen
    listener = PCMListener(gen_queue, pipeline)
    # speak
    speaker = TTSSpeaker(TTSSpeakerArgs(), gen_queue, pipeline, tts)

    inference_stream(listener, speaker, configs)


# format: ruff format bin/inference_stream.py



import torch
import torch.nn as nn
import torch.optim as optim
from torch.nn.parallel import DataParallel
import torch.utils.data as data
from torch.utils.data import DataLoader, TensorDataset

# 1. 定义一个简单的神经网络模型
class SimpleNN(nn.Module):
    def __init__(self, input_size, hidden_size, output_size):
        super(SimpleNN, self).__init__()
        self.fc1 = nn.Linear(input_size, hidden_size)
        self.relu = nn.ReLU()
        self.fc2 = nn.Linear(hidden_size, output_size)

    def forward(self, x):
        x = self.fc1(x)
        x = self.relu(x)
        x = self.fc2(x)
        return x

# 2. 生成一些模拟数据
def generate_dummy_data(num_samples, input_size):
    X = torch.randn(num_samples, input_size)  # 输入数据
    y = torch.randint(0, 2, (num_samples,))  # 随机标签（二分类）
    return X, y

# 3. 主函数
def train_data_parallel():
    # 参数设置
    input_size = 10
    hidden_size = 20
    output_size = 2
    batch_size = 32
    num_samples = 1000
    num_epochs = 5

    # 检查可用 GPU
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    print(f"Using device: {device}, GPU count: {torch.cuda.device_count()}")

    # 生成数据
    X, y = generate_dummy_data(num_samples, input_size)
    dataset = TensorDataset(X, y)
    dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=True)

    # 初始化模型
    model = SimpleNN(input_size, hidden_size, output_size)

    # 如果有多个 GPU，使用 DataParallel 包装模型 ， 单节点多卡
    if torch.cuda.device_count() > 1:
        print(f"Using {torch.cuda.device_count()} GPUs for Data Parallelism!")
        model = DataParallel(model)
    
    # 将模型移动到 GPU
    model = model.to(device)

    # 定义损失函数和优化器
    criterion = nn.CrossEntropyLoss()
    optimizer = optim.SGD(model.parameters(), lr=0.01)

    # 训练循环
    for epoch in range(num_epochs):
        model.train()
        running_loss = 0.0
        for i, (inputs, labels) in enumerate(dataloader):
            # 将数据移动到 GPU
            inputs, labels = inputs.to(device), labels.to(device)

            # 清零梯度
            optimizer.zero_grad()

            # 前向传播
            outputs = model(inputs)
            loss = criterion(outputs, labels)

            # 反向传播和优化
            loss.backward()
            optimizer.step()

            running_loss += loss.item()

        avg_loss = running_loss / len(dataloader)
        print(f"Epoch [{epoch+1}/{num_epochs}], Loss: {avg_loss:.4f}")

    print("Training finished!")

# 4. 运行训练
# python deploy/modal/src/train/demo/dp.py
if __name__ == "__main__":
    train_data_parallel()


import logging
from typing import AsyncGenerator

from apipeline.frames.data_frames import Frame
from apipeline.frames.sys_frames import ErrorFrame

from achatbot.common.factory import EngineClass
from achatbot.common.session import Session
from achatbot.common.utils.time import time_now_iso8601
from achatbot.common.interface import IAsr
from achatbot.processors.speech.asr.base import SegmentedASRProcessor
from achatbot.types.frames.data_frames import TranscriptionFrame
from achatbot.types.speech.language import Language


class ASRProcessor(SegmentedASRProcessor):
    def __init__(
        self,
        *,
        min_volume: float = 0.6,
        max_silence_secs: float = 0.3,
        max_buffer_secs: float = 1.5,
        sample_rate: int = 16000,
        num_channels: int = 1,
        asr: IAsr | EngineClass | None = None,
        session: Session | None = None,
        **kwargs,
    ):
        super().__init__(
            min_volume=min_volume,
            max_silence_secs=max_silence_secs,
            max_buffer_secs=max_buffer_secs,
            sample_rate=sample_rate,
            num_channels=num_channels,
            **kwargs,
        )
        self._asr = asr
        self._session = session

    def set_asr(self, asr: IAsr):
        self._asr = asr

    async def set_asr_args(self, **args):
        self._asr.set_args(**args)

    async def run_asr(self, audio: bytes) -> AsyncGenerator[Frame, None]:
        if self._asr is None:
            logging.error(f"{self} error: ASR engine not available")
            yield ErrorFrame("ASR engine not available")
            return

        await self.start_processing_metrics()
        await self.start_ttfb_metrics()

        self._asr.set_audio_data(audio)
        text: str = ""
        async for segment in self._asr.transcribe_stream(self._session):
            text += f"{segment}"

        await self.stop_ttfb_metrics()
        await self.stop_processing_metrics()

        language = None
        args = self._asr.get_args_dict()
        if "language" in args:
            language = Language(args["language"])

        if text:
            logging.info(f"{self._asr.SELECTED_TAG} Transcription: [{text}]")
            yield TranscriptionFrame(text, "", time_now_iso8601(), language)



from dataclasses import dataclass, field
import os

from src.common.types import MODELS_DIR


@dataclass
class KokoroTTSArgs:
    device: str = None
    ckpt_path: str = os.path.join(MODELS_DIR, "Kokoro82M/kokoro-v0_19.pth")
    voices_stats_dir: str = os.path.join(MODELS_DIR, "Kokoro82M/voices")  # for load all voices
    voice: str = "af"  # Default voice is a 50-50 mix of Bella & Sarah

    # language is determined by the first letter of the VOICE_NAME:
    # 🇺🇸 'a' => American English => en-us
    # 🇬🇧 'b' => British English => en-gb
    # so kokoro-v0_19 just support English
    language: str = "a"

    # generate params
    speed: float = 1.0

    # stream
    tts_stream: bool = False
    chunk_length_seconds: int = 1

    # silence
    add_silence_chunk: bool = False


@dataclass
class KokoroOnnxTTSArgs:
    # device: str = None
    model_struct_stats_ckpt: str = os.path.join(MODELS_DIR, "Kokoro82M/kokoro-v0_19.onnx")
    voices_file_path: str = os.path.join(MODELS_DIR, "Kokoro82M/kokoro-voices.json")
    espeak_ng_data_path: str = None
    espeak_ng_lib_path: str = None  # "/usr/local/lib/libespeak-ng.1.dylib"
    voice: str = "af"  # Default voice is a 50-50 mix of Bella & Sarah

    # # https://github.com/espeak-ng/espeak-ng/blob/master/docs/languages.md
    # "en-us",  # English
    # "en-gb",  # English (British)
    # "fr-fr",  # French
    # "ja",  # Japanese
    # "ko",  # Korean
    # "cmn",  # Mandarin Chinese
    language: str = "en-us"

    # generate params
    speed: float = 1.0

    # stream
    tts_stream: bool = False
    chunk_length_seconds: int = 1

    # silence
    add_silence_chunk: bool = False



# Copyright (c) 2024 Alibaba Inc (authors: Xiang Lyu)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import random
import json
import tarfile
import json
import io
import pyarrow.parquet as pq
from io import BytesIO
import torch
import torchaudio
from torch.nn.utils.rnn import pad_sequence
import torch.nn.functional as F
import tarfile
import json
import io
import wave
import numpy as np
import torchaudio
import os
import sys
import json
import random
import pickle
import argparse
import itertools
import mmap
import struct
import collections



import shutil
import multiprocessing as mp
from pathlib import Path

from tqdm import tqdm
from collections import defaultdict
from copy import deepcopy
from datetime import datetime
import pickle

from wids import wids
import math

torchaudio.set_audio_backend('soundfile')

AUDIO_FORMAT_SETS = set(['flac', 'mp3', 'm4a', 'ogg', 'opus', 'wav', 'wma'])

try:
    MAIN_SPK_EMBEDDING=torch.load("/workspace/audio_checkpoints/flow_model/spk_embedding/0909/mean_embedding.pt")
    GPT_SPK_EMBEDDING=torch.load("/workspace/audio_checkpoints/flow_model/spk_embedding/0909/spk_mean_embeddings.pt")
except:
    MAIN_SPK_EMBEDDING=torch.zeros(1,192)
    GPT_SPK_EMBEDDING=torch.zeros(1,192)

def parquet_opener(data, mode='train', tts_data={}):
    """ Give url or local file, return file descriptor
        Inplace operation.

        Args:
            data(Iterable[str]): url or local file list

        Returns:
            Iterable[{src, stream}]
    """
    for sample in data:
        assert 'src' in sample
        url = sample['src']
        try:
            df = pq.read_table(url).to_pandas()
            for i in range(len(df)):
                if mode == 'inference' and df.loc[i, 'utt'] not in tts_data:
                    continue
                sample.update(dict(df.loc[i]))
                if mode == 'train':
                    # NOTE do not return sample directly, must initialize a new dict
                    yield {**sample}
                else:
                    for index, text in enumerate(tts_data[df.loc[i, 'utt']]):
                        yield {**sample, 'tts_index': index, 'tts_text': text}
        except Exception as ex:
            logging.warning('Failed to open {}, ex info {}'.format(url, ex))




def parse_tar_header(header_bytes):
    header = struct.unpack("!100s8s8s8s12s12s8s1s100s6s2s32s32s8s8s155s", header_bytes)
    return TarHeader(*header)

TarHeader = collections.namedtuple(
    "TarHeader",
    [
        "name",
        "mode",
        "uid",
        "gid",
        "size",
        "mtime",
        "chksum",
        "typeflag",
        "linkname",
        "magic",
        "version",
        "uname",
        "gname",
        "devmajor",
        "devminor",
        "prefix",
    ],
)

class MMTar:
    def __init__(self, file_path: Path | str):
        self.stream = open(file_path, "rb")
        self.mmap = mmap.mmap(self.stream.fileno(), 0, access=mmap.ACCESS_READ)

    def __del__(self):
        try:
            self.mmap.close()
            self.stream.close()
        except:  # noqa
            pass

    def get_at_offset(self, offset) -> tuple[str, bytes]:
        header = parse_tar_header(self.mmap[offset : offset + 500])
        name = header.name.decode("utf-8").strip("\x00")
        start = offset + 512
        end = start + int(header.size.decode("utf-8")[:-1], 8)
        return name, self.mmap[start:end]


class Tar:
    def __init__(self, path: Path):
        self.tar = MMTar(path)
        indices_path = path.with_suffix(".index")
        self.index = pickle.loads(indices_path.read_bytes())
        self.name_mapping = {}
        for name, offset, _ in self.index:
            self.name_mapping[name] = offset

    def read(self, name: str) -> bytes:
        return self.tar.get_at_offset(self.name_mapping[name])[1]

def cosy_jsonl_opener(data, mode='train', tts_data={}):
    """ Give url or local file, return file descriptor
        Inplace operation.

        Args:
            data(Iterable[str]): url or local file list

        Returns:
            Iterable[{src, stream}]
    """
    for sample in data:
        assert 'src' in sample
        cosy_jsonl_path = sample['src']
        tar_file_path=cosy_jsonl_path.replace(".vq0907.jsonl",".tar")
        try:
            tar_data=Tar(Path(tar_file_path))
            with open(cosy_jsonl_path, 'r') as f:
                for line in f:
                    item=json.loads(line)
                    cosy_token = item['cosy_token']
                    sample['speech_token']=torch.tensor(cosy_token)
                    sample['speech'], sample['sample_rate']= torchaudio.load(io.BytesIO(tar_data.read(item['filename'])))
                    # print(item['filename'])
                    yield {**sample}

        except Exception as ex:
            logging.warning('Failed to open {}, ex info {}'.format(cosy_jsonl_path, ex))


def cosy_jsonl_opener_vq0918_nopool(data, mode='train', tts_data={}):
    """ Give url or local file, return file descriptor
        Inplace operation.

        Args:
            data(Iterable[str]): url or local file list

        Returns:
            Iterable[{src, stream}]
    """
    for sample in data:
        assert 'src' in sample
        cosy_jsonl_path = sample['src']
        tar_file_path=cosy_jsonl_path.replace(".vq0918-nopool.jsonl",".tar")


        try:
            tar_data=Tar(Path(tar_file_path))
            with open(cosy_jsonl_path, 'r') as f:
                # cosy_data = [json.loads(line) for line in f]
                for line in f:
                    item=json.loads(line)
                    cosy_token = item['cosy_token']
                    sample['speech_token']=torch.tensor(cosy_token)
                    sample['speech'], sample['sample_rate']= torchaudio.load(io.BytesIO(tar_data.read(item['filename'])))
                    # print(item['filename'])
                    yield {**sample}

        except Exception as ex:
            logging.warning('Failed to open {}, ex info {}'.format(cosy_jsonl_path, ex))



def cosy_jsonl_opener_vq0918_pool2(data, mode='train', tts_data={}):
    """ Give url or local file, return file descriptor
        Inplace operation.

        Args:
            data(Iterable[str]): url or local file list

        Returns:
            Iterable[{src, stream}]
    """
    for sample in data:
        assert 'src' in sample
        cosy_jsonl_path = sample['src']
        tar_file_path=cosy_jsonl_path.replace(".vq0918-pool2.jsonl",".tar")

        try:
            tar_data=Tar(Path(tar_file_path))
            with open(cosy_jsonl_path, 'r') as f:
                for line in f:
                    item=json.loads(line)
                    cosy_token = item['cosy_token']
                    sample['speech_token']=torch.tensor(cosy_token)
                    sample['speech'], sample['sample_rate']= torchaudio.load(io.BytesIO(tar_data.read(item['filename'])))

                    yield {**sample}

        except Exception as ex:
            logging.warning('Failed to open {}, ex info {}'.format(cosy_jsonl_path, ex))


def cosy_jsonl_opener_vq0918_pool4(data, mode='train', tts_data={}):
    """ Give url or local file, return file descriptor
        Inplace operation.

        Args:
            data(Iterable[str]): url or local file list

        Returns:
            Iterable[{src, stream}]
    """
    for sample in data:
        assert 'src' in sample
        cosy_jsonl_path = sample['src']
        tar_file_path=cosy_jsonl_path.replace(".vq0918-pool4.jsonl",".tar")
        try:
            tar_data=Tar(Path(tar_file_path))
            with open(cosy_jsonl_path, 'r') as f:
                # cosy_data = [json.loads(line) for line in f]
                for line in f:
                    item=json.loads(line)
                    cosy_token = item['cosy_token']
                    sample['speech_token']=torch.tensor(cosy_token)
                    sample['speech'], sample['sample_rate']= torchaudio.load(io.BytesIO(tar_data.read(item['filename'])))
                    # print(item['filename'])
                    yield {**sample}

        except Exception as ex:
            logging.warning('Failed to open {}, ex info {}'.format(cosy_jsonl_path, ex))


def cosy_jsonl_opener_vq0918_pool8(data, mode='train', tts_data={}):
    """ Give url or local file, return file descriptor
        Inplace operation.

        Args:
            data(Iterable[str]): url or local file list

        Returns:
            Iterable[{src, stream}]
    """
    for sample in data:
        assert 'src' in sample
        cosy_jsonl_path = sample['src']
        tar_file_path=cosy_jsonl_path.replace(".vq0918-pool8.jsonl",".tar")

        try:
            tar_data=Tar(Path(tar_file_path))
            with open(cosy_jsonl_path, 'r') as f:
                # cosy_data = [json.loads(line) for line in f]
                for line in f:
                    item=json.loads(line)
                    cosy_token = item['cosy_token']
                    sample['speech_token']=torch.tensor(cosy_token)
                    sample['speech'], sample['sample_rate']= torchaudio.load(io.BytesIO(tar_data.read(item['filename'])))
                    # print(item['filename'])
                    yield {**sample}

        except Exception as ex:
            logging.warning('Failed to open {}, ex info {}'.format(cosy_jsonl_path, ex))
         


def process_sft_vq0918_pool4(data, mode='train', tts_data={}):
    for sample in data:
        assert 'src' in sample
        
        token_npy_path = sample['src']
        wav_path=token_npy_path.replace(".vq0918-pool4.npy","")

        # wav_path,token_npy_path=sample['src'].split(' ')
        try:
            sample['speech_token']=torch.tensor(np.load(token_npy_path))
            sample['speech'], sample['sample_rate']= torchaudio.load(wav_path)
            if sample['speech'].shape[0] > 1:
                sample['speech'] = sample['speech'].mean(dim=0, keepdim=True)
            sample['spk_embedding']=torch.zeros_like(MAIN_SPK_EMBEDDING)
            yield {**sample}
        except Exception as ex:
            logging.warning('Failed to open {}, ex info {}'.format(wav_path, ex))
            logging.warning('Failed to open {}'.format(wav_path))


def process_sft_vq0918_pool4_split(data, mode='train',split_token=25, tts_data={}):
    for sample in data:
        assert 'src' in sample
        
        token_npy_path = sample['src']
        wav_path=token_npy_path.replace(".vq0918-pool4.npy","")

        # wav_path,token_npy_path=sample['src'].split(' ')
        try:
            # sample['speech_token']=torch.tensor(np.load(token_npy_path))
            # sample['speech'], sample['sample_rate']= torchaudio.load(wav_path)
            # if sample['speech'].shape[0] > 1:
            #     sample['speech'] = sample['speech'].mean(dim=0, keepdim=True)
            
            
            # sample['spk_embedding']=torch.zeros_like(MAIN_SPK_EMBEDDING)


            speech_token=torch.tensor(np.load(token_npy_path))
            speech,sample_rate= torchaudio.load(wav_path)
            # split_speech=int(split_token / 12.5 * sample_rate)
            if speech.shape[0] > 1:
                speech = speech.mean(dim=0, keepdim=True)

            sample['spk_embedding']=torch.zeros_like(MAIN_SPK_EMBEDDING)
            sample['sample_rate']=sample_rate

            num_splits = (speech_token.size(0) + split_token - 1) // split_token 

            for split_id in range(num_splits):
                end_token_idx = min((split_id + 1) * split_token, speech_token.size(0))
                end_speech_idx=int(np.ceil(end_token_idx / 12.5 * sample_rate))
                sample['speech_token']=speech_token[:end_token_idx]
                sample['speech']=speech[:,:end_speech_idx]
                print(sample['speech_token'].size(),sample['speech'].size())
                yield {**sample}
        except Exception as ex:
            logging.warning('Failed to open {}, ex info {}'.format(wav_path, ex))
            logging.warning('Failed to open {}'.format(wav_path))


def process_sft_vq0918_pool2(data, mode='train', tts_data={}):
    for sample in data:
        assert 'src' in sample
        
        token_npy_path = sample['src'].replace(".vq0918-pool4.npy",".vq0918-pool2.npy")
        wav_path=token_npy_path.replace(".vq0918-pool2.npy","")

        # wav_path,token_npy_path=sample['src'].split(' ')
        try:
            sample['speech_token']=torch.tensor(np.load(token_npy_path))
            sample['speech'], sample['sample_rate']= torchaudio.load(wav_path)
            if sample['speech'].shape[0] > 1:
                sample['speech'] = sample['speech'].mean(dim=0, keepdim=True)

            sample['spk_embedding']=torch.zeros_like(MAIN_SPK_EMBEDDING)
            yield {**sample}
        except Exception as ex:
            logging.warning('Failed to open {}, ex info {}'.format(wav_path, ex))
            logging.warning('Failed to open {}'.format(wav_path))
        

def process_sft_vq0918_pool2_split(data, mode='train',split_token=50, tts_data={}):
    for sample in data:
        assert 'src' in sample
        
        token_npy_path = sample['src']
        wav_path=token_npy_path.replace(".vq0918-pool2.npy","")

        # wav_path,token_npy_path=sample['src'].split(' ')
        try:
            # sample['speech_token']=torch.tensor(np.load(token_npy_path))
            # sample['speech'], sample['sample_rate']= torchaudio.load(wav_path)
            # if sample['speech'].shape[0] > 1:
            #     sample['speech'] = sample['speech'].mean(dim=0, keepdim=True)
            
            
            # sample['spk_embedding']=torch.zeros_like(MAIN_SPK_EMBEDDING)


            speech_token=torch.tensor(np.load(token_npy_path))
            speech,sample_rate= torchaudio.load(wav_path)
            # split_speech=int(split_token / 12.5 * sample_rate)
            if speech.shape[0] > 1:
                speech = speech.mean(dim=0, keepdim=True)

            sample['spk_embedding']=torch.zeros_like(MAIN_SPK_EMBEDDING)
            sample['sample_rate']=sample_rate

            num_splits = (speech_token.size(0) + split_token - 1) // split_token 

            for split_id in range(num_splits):
                end_token_idx = min((split_id + 1) * split_token, speech_token.size(0))
                end_speech_idx=int(np.ceil(end_token_idx / 25 * sample_rate))
                sample['speech_token']=speech_token[:end_token_idx]
                sample['speech']=speech[:,:end_speech_idx]
                print(sample['speech_token'].size(),sample['speech'].size())
                yield {**sample}
        except Exception as ex:
            logging.warning('Failed to open {}, ex info {}'.format(wav_path, ex))
            logging.warning('Failed to open {}'.format(wav_path))

def process_sft_vq0918_pool4_gpt(data, mode='train', tts_data={}):
    for sample in data:
        assert 'src' in sample
        try:
            entry=json.loads(sample['src'])
            sample['spk_embedding']=torch.zeros_like(MAIN_SPK_EMBEDDING)
            
            for conv in entry["conversations"]:
                if "response_wav" in conv:
                    wav_path = f"/workspace/audio_data/sft/{conv['response_wav']}"
                    token_npy_path=wav_path.replace(".wav",".wav.vq0918-pool4.npy")
                    sample['speech_token']=torch.tensor(np.load(token_npy_path))
                    sample['speech'], sample['sample_rate']= torchaudio.load(wav_path)
                    if sample['speech'].shape[0] > 1:
                        sample['speech'] = sample['speech'].mean(dim=0, keepdim=True)
                    sample['spk_embedding']=spk_embedding
                    yield {**sample}
        except Exception as ex:
            # logging.warning('Failed to open {}, ex info {}'.format(wav_path, ex))
            logging.warning('Failed to open {}'.format(wav_path))


def process_sft_vq0918_pool4_gpt_1010(data, mode='train', tts_data={}):
    for sample in data:
        assert 'src' in sample
        try:
            entry=json.loads(sample['src'])
            sample['spk_embedding']=torch.zeros_like(MAIN_SPK_EMBEDDING)
            
            for conv in entry["conversations"]:
                if "response_wav" in conv:
                    wav_path = f"/workspace/audio_data/sft/{conv['response_wav']}"
                    token_npy_path=wav_path.replace(".wav",".wav.vq0918-pool4.npy")
                    sample['speech_token']=torch.tensor(np.load(token_npy_path))
                    sample['speech'], sample['sample_rate']= torchaudio.load(wav_path)
                    if sample['speech'].shape[0] > 1:
                        sample['speech'] = sample['speech'].mean(dim=0, keepdim=True)
                    sample['spk_embedding']=spk_embedding
                    yield {**sample}
                if "prompt_wav" in conv:
                    wav_path = f"/workspace/audio_data/sft/{conv['response_wav']}"
                    token_npy_path=wav_path.replace(".wav",".wav.vq0918-pool4.npy")
                    sample['speech_token']=torch.tensor(np.load(token_npy_path))
                    sample['speech'], sample['sample_rate']= torchaudio.load(wav_path)
                    if sample['speech'].shape[0] > 1:
                        sample['speech'] = sample['speech'].mean(dim=0, keepdim=True)
                    sample['spk_embedding']=spk_embedding
                    yield {**sample}
        except Exception as ex:
            # logging.warning('Failed to open {}, ex info {}'.format(wav_path, ex))
            logging.warning('Failed to open {}'.format(wav_path))


def filter(data,
           max_length=10240,
           min_length=10,
           token_max_length=200,
           token_min_length=1,
           min_output_input_ratio=0.0005,
           max_output_input_ratio=1,
           mode='train'):
    """ Filter sample according to feature and label length
        Inplace operation.

        Args::
            data: Iterable[{key, wav, label, sample_rate}]
            max_length: drop utterance which is greater than max_length(10ms)
            min_length: drop utterance which is less than min_length(10ms)
            token_max_length: drop utterance which is greater than
                token_max_length, especially when use char unit for
                english modeling
            token_min_length: drop utterance which is
                less than token_max_length
            min_output_input_ratio: minimal ration of
                token_length / feats_length(10ms)
            max_output_input_ratio: maximum ration of
                token_length / feats_length(10ms)

        Returns:
            Iterable[{key, wav, label, sample_rate}]
    """
    for sample in data:
        # sample['speech'], sample['sample_rate'] = torchaudio.load(BytesIO(sample['audio_data']))
        # del sample['audio_data']
        # sample['wav'] is torch.Tensor, we have 100 frames every second
        num_frames = sample['speech'].size(1) / sample['sample_rate'] * 100
        if num_frames < min_length:
            continue
        if num_frames > max_length:
            continue
        if len(sample['text_token']) < token_min_length:
            continue
        if len(sample['text_token']) > token_max_length:
            continue
        if len(sample['speech_token']) == 0:
            continue
        if num_frames != 0:
            if len(sample['text_token']) / num_frames < min_output_input_ratio:
                continue
            if len(sample['text_token']) / num_frames > max_output_input_ratio:
                continue
        yield sample

            
def filter_speech_token(data,
           max_length=10240,
           min_length=10,
           token_max_length=5000,
           token_min_length=1,
           min_output_input_ratio=0.0005,
           max_output_input_ratio=30,
           mode='train'):
    """ Filter sample according to feature and label length
        Inplace operation.

        Args::
            data: Iterable[{key, wav, label, sample_rate}]
            max_length: drop utterance which is greater than max_length(10ms)
            min_length: drop utterance which is less than min_length(10ms)
            token_max_length: drop utterance which is greater than
                token_max_length, especially when use char unit for
                english modeling
            token_min_length: drop utterance which is
                less than token_max_length
            min_output_input_ratio: minimal ration of
                token_length / feats_length(10ms)
            max_output_input_ratio: maximum ration of
                token_length / feats_length(10ms)

        Returns:
            Iterable[{key, wav, label, sample_rate}]
    """
    for sample in data:
        # sample['speech'], sample['sample_rate'] = torchaudio.load(BytesIO(sample['audio_data']))
        # del sample['audio_data']
        # sample['wav'] is torch.Tensor, we have 100 frames every second
        num_frames = sample['speech'].size(1) / sample['sample_rate'] * 100
        if num_frames < min_length:
            continue
        if num_frames > max_length:
            continue
        if len(sample['speech_token']) < token_min_length:
            continue
        if len(sample['speech_token']) > token_max_length:
            continue
        if len(sample['speech_token']) == 0:
            continue
        if num_frames != 0:
            if len(sample['speech_token']) / num_frames < min_output_input_ratio:
                continue
            if len(sample['speech_token']) / num_frames > max_output_input_ratio:
                continue
        yield sample


def resample(data, resample_rate=22050, min_sample_rate=16000, mode='train'):
    """ Resample data.
        Inplace operation.

        Args:
            data: Iterable[{key, wav, label, sample_rate}]
            resample_rate: target resample rate

        Returns:
            Iterable[{key, wav, label, sample_rate}]
    """
    for sample in data:
        assert 'sample_rate' in sample
        assert 'speech' in sample
        sample_rate = sample['sample_rate']
        waveform = sample['speech']
        if sample_rate != resample_rate:
            if sample_rate < min_sample_rate:
                continue
            sample['sample_rate'] = resample_rate
            sample['speech'] = torchaudio.transforms.Resample(
                orig_freq=sample_rate, new_freq=resample_rate)(waveform)
        max_val = sample['speech'].abs().max()
        if max_val > 1:
            sample['speech'] /= max_val
        yield sample


def compute_fbank(data,
                  feat_extractor,
                  mode='train'):
    """ Extract fbank

        Args:
            data: Iterable[{key, wav, label, sample_rate}]

        Returns:
            Iterable[{key, feat, label}]
    """
    for sample in data:
        assert 'sample_rate' in sample
        assert 'speech' in sample
        # assert 'utt' in sample
        # assert 'text_token' in sample
        waveform = sample['speech']
        mat = feat_extractor(waveform).squeeze(dim=0).transpose(0, 1)
        sample['speech_feat'] = mat
        del sample['speech']
        yield sample


def parse_embedding(data, normalize, mode='train'):
    """ Parse utt_embedding/spk_embedding

        Args:
            data: Iterable[{key, wav, label, sample_rate}]

        Returns:
            Iterable[{key, feat, label}]
    """
    for sample in data:
        sample['utt_embedding'] = torch.tensor(sample['utt_embedding'], dtype=torch.float32)
        sample['spk_embedding'] = torch.tensor(sample['spk_embedding'], dtype=torch.float32)
        if normalize:
            sample['utt_embedding'] = F.normalize(sample['utt_embedding'], dim=0)
            sample['spk_embedding'] = F.normalize(sample['spk_embedding'], dim=0)
        yield sample


def tokenize(data, get_tokenizer, allowed_special, mode='train'):
    """ Decode text to chars or BPE
        Inplace operation

        Args:
            data: Iterable[{key, wav, txt, sample_rate}]

        Returns:
            Iterable[{key, wav, txt, tokens, label, sample_rate}]
    """
    tokenizer = get_tokenizer()
    for sample in data:
        assert 'text' in sample
        sample['text_token'] = tokenizer.encode(sample['text'], allowed_special=allowed_special)
        if mode == 'inference':
            sample['tts_text_token'] = tokenizer.encode(sample['tts_text'], allowed_special=allowed_special)
        yield sample


def shuffle(data, shuffle_size=10000, mode='train'):
    """ Local shuffle the data

        Args:
            data: Iterable[{key, feat, label}]
            shuffle_size: buffer size for shuffle

        Returns:
            Iterable[{key, feat, label}]
    """
    buf = []
    for sample in data:
        buf.append(sample)
        if len(buf) >= shuffle_size:
            random.shuffle(buf)
            for x in buf:
                yield x
            buf = []
    # The sample left over
    random.shuffle(buf)
    for x in buf:
        yield x


def sort(data, sort_size=500, mode='train'):
    """ Sort the data by feature length.
        Sort is used after shuffle and before batch, so we can group
        utts with similar lengths into a batch, and `sort_size` should
        be less than `shuffle_size`

        Args:
            data: Iterable[{key, feat, label}]
            sort_size: buffer size for sort

        Returns:
            Iterable[{key, feat, label}]
    """

    buf = []
    for sample in data:
        buf.append(sample)
        if len(buf) >= sort_size:
            buf.sort(key=lambda x: x['speech_feat'].size(0))
            for x in buf:
                yield x
            buf = []
    # The sample left over
    buf.sort(key=lambda x: x['speech_feat'].size(0))
    for x in buf:
        yield x


def static_batch(data, batch_size=16):
    """ Static batch the data by `batch_size`

        Args:
            data: Iterable[{key, feat, label}]
            batch_size: batch size

        Returns:
            Iterable[List[{key, feat, label}]]
    """
    buf = []
    for sample in data:
        buf.append(sample)
        if len(buf) >= batch_size:
            yield buf
            buf = []
    if len(buf) > 0:
        yield buf


def dynamic_batch(data, max_frames_in_batch=12000, mode='train'):
    """ Dynamic batch the data until the total frames in batch
        reach `max_frames_in_batch`

        Args:
            data: Iterable[{key, feat, label}]
            max_frames_in_batch: max_frames in one batch

        Returns:
            Iterable[List[{key, feat, label}]]
    """
    buf = []
    longest_frames = 0
    for sample in data:
        assert 'speech_feat' in sample
        assert isinstance(sample['speech_feat'], torch.Tensor)
        new_sample_frames = sample['speech_feat'].size(0)
        longest_frames = max(longest_frames, new_sample_frames)
        frames_after_padding = longest_frames * (len(buf) + 1)
        if frames_after_padding > max_frames_in_batch:
            yield buf
            buf = [sample]
            longest_frames = new_sample_frames
        else:
            buf.append(sample)
    if len(buf) > 0:
        yield buf


def batch(data, batch_type='static', batch_size=16, max_frames_in_batch=12000, mode='train'):
    """ Wrapper for static/dynamic batch
    """
    if mode == 'inference':
        return static_batch(data, 1)
    else:
        if batch_type == 'static':
            return static_batch(data, batch_size)
        elif batch_type == 'dynamic':
            return dynamic_batch(data, max_frames_in_batch)
        else:
            logging.fatal('Unsupported batch type {}'.format(batch_type))


def padding(data, use_spk_embedding, mode='train'):
    """ Padding the data into training data

        Args:
            data: Iterable[List[{key, feat, label}]]

        Returns:
            Iterable[Tuple(keys, feats, labels, feats lengths, label lengths)]
    """
    for sample in data:
        assert isinstance(sample, list)
        speech_feat_len = torch.tensor([x['speech_feat'].size(1) for x in sample],
                                       dtype=torch.int32)
        order = torch.argsort(speech_feat_len, descending=True)

        utts = [sample[i]['utt'] for i in order]
        speech_token = [torch.tensor(sample[i]['speech_token']) for i in order]
        speech_token_len = torch.tensor([i.size(0) for i in speech_token], dtype=torch.int32)
        speech_token = pad_sequence(speech_token,
                                    batch_first=True,
                                    padding_value=0)
        speech_feat = [sample[i]['speech_feat'] for i in order]
        speech_feat_len = torch.tensor([i.size(0) for i in speech_feat], dtype=torch.int32)
        speech_feat = pad_sequence(speech_feat,
                                   batch_first=True,
                                   padding_value=0)
        text = [sample[i]['text'] for i in order]
        text_token = [torch.tensor(sample[i]['text_token']) for i in order]
        text_token_len = torch.tensor([i.size(0) for i in text_token], dtype=torch.int32)
        text_token = pad_sequence(text_token, batch_first=True, padding_value=0)
        utt_embedding = torch.stack([sample[i]['utt_embedding'] for i in order], dim=0)
        spk_embedding = torch.stack([sample[i]['spk_embedding'] for i in order], dim=0)
        batch = {
            "utts": utts,
            "speech_token": speech_token,
            "speech_token_len": speech_token_len,
            "speech_feat": speech_feat,
            "speech_feat_len": speech_feat_len,
            "text": text,
            "text_token": text_token,
            "text_token_len": text_token_len,
            "utt_embedding": utt_embedding,
            "spk_embedding": spk_embedding,
        }
        if mode == 'inference':
            tts_text = [sample[i]['tts_text'] for i in order]
            tts_index = [sample[i]['tts_index'] for i in order]
            tts_text_token = [torch.tensor(sample[i]['tts_text_token']) for i in order]
            tts_text_token_len = torch.tensor([i.size(0) for i in tts_text_token], dtype=torch.int32)
            tts_text_token = pad_sequence(tts_text_token, batch_first=True, padding_value=-1)
            batch.update({'tts_text': tts_text,
                          'tts_index': tts_index,
                          'tts_text_token': tts_text_token,
                          'tts_text_token_len': tts_text_token_len})
        if use_spk_embedding is True:
            batch["embedding"] = batch["spk_embedding"]
        else:
            batch["embedding"] = batch["utt_embedding"]
        yield batch



def padding_speech_token(data, use_spk_embedding, mode='train'):
    """ Padding the data into training data

        Args:
            data: Iterable[List[{key, feat, label}]]

        Returns:
            Iterable[Tuple(keys, feats, labels, feats lengths, label lengths)]
    """
    for sample in data:
        assert isinstance(sample, list)
        speech_feat_len = torch.tensor([x['speech_feat'].size(1) for x in sample],
                                       dtype=torch.int32)
        order = torch.argsort(speech_feat_len, descending=True)

        # utts = [sample[i]['utt'] for i in order]
        # speech_token = [torch.tensor(sample[i]['speech_token']) for i in order]
        try:
            speech_token = [sample[i]['speech_token'].clone().detach() for i in order]
            speech_token_len = torch.tensor([i.size(0) for i in speech_token], dtype=torch.int32)
            speech_token = pad_sequence(speech_token,
                                        batch_first=True,
                                        padding_value=0)
            speech_feat = [sample[i]['speech_feat'] for i in order]
            speech_feat_len = torch.tensor([i.size(0) for i in speech_feat], dtype=torch.int32)
            speech_feat = pad_sequence(speech_feat,
                                    batch_first=True,
                                    padding_value=0)
            batch = {
                "speech_token": speech_token,
                "speech_token_len": speech_token_len,
                "speech_feat": speech_feat,
                "speech_feat_len": speech_feat_len,
            }
            if mode == 'inference':
                tts_text = [sample[i]['tts_text'] for i in order]
                tts_index = [sample[i]['tts_index'] for i in order]
                tts_text_token = [torch.tensor(sample[i]['tts_text_token']) for i in order]
                tts_text_token_len = torch.tensor([i.size(0) for i in tts_text_token], dtype=torch.int32)
                tts_text_token = pad_sequence(tts_text_token, batch_first=True, padding_value=-1)
                batch.update({'tts_text': tts_text,
                            'tts_index': tts_index,
                            'tts_text_token': tts_text_token,
                            'tts_text_token_len': tts_text_token_len})
            # if use_spk_embedding is True:
            #     batch["embedding"] = batch["spk_embedding"]
            # else:
            #     batch["embedding"] = batch["utt_embedding"]
            batch["embedding"]=torch.zeros((batch["speech_feat"].size(0),192),device=batch["speech_feat"].device)
            yield batch
        except Exception as ex:
            logging.warning(' ex info {}'.format(ex))
            # assert False



def padding_speech_token_spk(data, use_spk_embedding, mode='train'):
    """ Padding the data into training data

        Args:
            data: Iterable[List[{key, feat, label}]]

        Returns:
            Iterable[Tuple(keys, feats, labels, feats lengths, label lengths)]
    """
    for sample in data:
        assert isinstance(sample, list)
        speech_feat_len = torch.tensor([x['speech_feat'].size(1) for x in sample],
                                       dtype=torch.int32)
        order = torch.argsort(speech_feat_len, descending=True)

        # utts = [sample[i]['utt'] for i in order]
        # speech_token = [torch.tensor(sample[i]['speech_token']) for i in order]
        try:
            speech_token = [sample[i]['speech_token'].clone().detach() for i in order]
            speech_token_len = torch.tensor([i.size(0) for i in speech_token], dtype=torch.int32)
            speech_token = pad_sequence(speech_token,
                                        batch_first=True,
                                        padding_value=0)
            speech_feat = [sample[i]['speech_feat'] for i in order]
            speech_feat_len = torch.tensor([i.size(0) for i in speech_feat], dtype=torch.int32)
            speech_feat = pad_sequence(speech_feat,
                                    batch_first=True,
                                    padding_value=0)
            spk_embedding = torch.stack([sample[i]['spk_embedding'] for i in order], dim=0)
            batch = {
                "speech_token": speech_token,
                "speech_token_len": speech_token_len,
                "speech_feat": speech_feat,
                "speech_feat_len": speech_feat_len,
                "spk_embedding": spk_embedding,
            }
            if mode == 'inference':
                tts_text = [sample[i]['tts_text'] for i in order]
                tts_index = [sample[i]['tts_index'] for i in order]
                tts_text_token = [torch.tensor(sample[i]['tts_text_token']) for i in order]
                tts_text_token_len = torch.tensor([i.size(0) for i in tts_text_token], dtype=torch.int32)
                tts_text_token = pad_sequence(tts_text_token, batch_first=True, padding_value=-1)
                batch.update({'tts_text': tts_text,
                            'tts_index': tts_index,
                            'tts_text_token': tts_text_token,
                            'tts_text_token_len': tts_text_token_len})
            # if use_spk_embedding is True:
            #     batch["embedding"] = batch["spk_embedding"]
            # else:
            #     batch["embedding"] = batch["utt_embedding"]
            # batch["embedding"]=torch.zeros((batch["speech_feat"].size(0),192),device=batch["speech_feat"].device)
            batch["embedding"] = batch["spk_embedding"]
            yield batch
        except Exception as ex:
            logging.warning(' ex info {}'.format(ex))
            # assert False


from dataclasses import dataclass, field

from sglang.srt.server_args import ServerArgs
from .sampling import LMGenerateArgs


@dataclass
class SGLangEngineArgs:
    """
    SGLang language model engine args
    """

    serv_args: dict = field(default_factory=lambda: ServerArgs().__dict__)
    gen_args: dict = field(default_factory=lambda: LMGenerateArgs().__dict__)



import io
import os
import logging

import unittest
from PIL import Image

import aiohttp
from apipeline.frames.sys_frames import Frame
from apipeline.pipeline.pipeline import Pipeline, FrameProcessor
from apipeline.pipeline.task import PipelineTask, PipelineParams
from apipeline.pipeline.runner import PipelineRunner
from apipeline.frames.data_frames import TextFrame, ImageRawFrame
from apipeline.processors.frame_processor import FrameDirection

from src.processors.image import get_image_gen_processor
from src.common.logger import Logger
from apipeline.processors.logger import FrameLogger

from dotenv import load_dotenv

load_dotenv()


"""
mkdir -p images

IMAGE_GEN_PROCESSOR=HFApiInferenceImageGenProcessor \
    python -m unittest test.integration.processors.test_image_gen_processor.TestProcessor

IMAGE_GEN_PROCESSOR=OpenAIImageGenProcessor \
    python -m unittest test.integration.processors.test_image_gen_processor.TestProcessor

IMAGE_GEN_PROCESSOR=TogetherImageGenProcessor \
    python -m unittest test.integration.processors.test_image_gen_processor.TestProcessor

IMAGE_GEN_PROCESSOR=HFStableDiffusionImageGenProcessor \
    python -m unittest test.integration.processors.test_image_gen_processor.TestProcessor
"""


class SaveImageProcessor(FrameProcessor):
    def __init__(self, save_file: str = "./images/test_gen.jpeg", **kwargs):
        super().__init__(**kwargs)
        self.save_file = save_file

    async def process_frame(self, frame: Frame, direction: FrameDirection):
        await super().process_frame(frame, direction)
        if isinstance(frame, ImageRawFrame):
            print("save-->", frame, self.save_file)
            img = Image.frombytes(mode=frame.mode, size=frame.size, data=frame.image)
            img.save(self.save_file, frame.format)


class TestProcessor(unittest.IsolatedAsyncioTestCase):
    @classmethod
    def setUpClass(cls):
        Logger.init(os.getenv("LOG_LEVEL", "info").upper(), is_file=False)
        cls.processor = os.getenv("IMAGE_GEN_PROCESSOR", "HFApiInferenceImageGenProcessor")
        cls.save_img_file = os.getenv("SAVE_FILE_PATH", "./images/test_gen.jpeg")
        cls.device = os.getenv("DEVICE", "auto")
        cls.client_session = None

    @classmethod
    def tearDownClass(cls):
        pass

    def get_image_gen_processor_by_tag(self):
        kwargs = {}
        if self.processor == "HFApiInferenceImageGenProcessor":
            self.client_session = aiohttp.ClientSession()
            kwargs["aiohttp_session"] = self.client_session
            kwargs["width"] = 1024
            kwargs["height"] = 1024
            kwargs["steps"] = 28
            kwargs["model"] = "stabilityai/stable-diffusion-3.5-large"
        if self.processor == "OpenAIImageGenProcessor":
            self.client_session = aiohttp.ClientSession()
            kwargs["aiohttp_session"] = self.client_session
            kwargs["image_size"] = "1024x1024"
            kwargs["model"] = "dall-e-3"
        if self.processor == "TogetherImageGenProcessor":
            kwargs["width"] = 1280
            kwargs["height"] = 720
            kwargs["steps"] = 4
            kwargs["model"] = "black-forest-labs/FLUX.1-schnell-Free"
        if self.processor == "HFStableDiffusionImageGenProcessor":
            kwargs["width"] = 1280
            kwargs["height"] = 720
            kwargs["steps"] = 28
            kwargs["guidance_scale"] = 3.5
            kwargs["is_quantizing"] = True
            kwargs["device"] = self.device
            kwargs["model"] = os.getenv("SD_MODEL", "stabilityai/stable-diffusion-3.5-large")

        return get_image_gen_processor(self.processor, **kwargs)

    def get_image_gen_processor(self):
        if self.processor == "HFApiInferenceImageGenProcessor":
            from src.processors.image.hf_img_gen_processor import HFApiInferenceImageGenProcessor

            self.client_session = aiohttp.ClientSession()
            return HFApiInferenceImageGenProcessor(
                aiohttp_session=self.client_session,
                model="stabilityai/stable-diffusion-3.5-large",
            )
        if self.processor == "OpenAIImageGenProcessor":
            from src.processors.image.openai_img_gen_processor import OpenAIImageGenProcessor

            self.client_session = aiohttp.ClientSession()
            return OpenAIImageGenProcessor(
                image_size="1024x1024",
                aiohttp_session=self.client_session,
                model="dall-e-3",
            )
        if self.processor == "TogetherImageGenProcessor":
            from src.processors.image.together_img_gen_processor import TogetherImageGenProcessor

            return TogetherImageGenProcessor(
                width=512,
                height=512,
                model="black-forest-labs/FLUX.1-schnell-Free",
            )

    async def asyncSetUp(self):
        pipeline = Pipeline(
            [
                self.get_image_gen_processor_by_tag(),
                FrameLogger(include_frame_types=[ImageRawFrame]),
                SaveImageProcessor(save_file=self.save_img_file),
            ]
        )

        self.task = PipelineTask(pipeline, PipelineParams())

    async def asyncTearDown(self):
        self.client_session and await self.client_session.close()
        pass

    async def test_run(self):
        runner = PipelineRunner()
        await self.task.queue_frames(
            [
                TextFrame(
                    text="A capybara holding a sign that reads Hello.",
                ),
            ]
        )
        await self.task.stop_when_done()
        await runner.run(self.task)



import logging

from apipeline.frames.sys_frames import MetricsFrame, CancelFrame
from apipeline.frames.data_frames import ImageRawFrame
from apipeline.frames.control_frames import StartFrame, EndFrame

from src.services.livekit_client import LivekitTransportClient
from src.processors.audio_camera_output_processor import AudioCameraOutputProcessor
from src.common.types import LivekitParams
from src.types.frames.data_frames import TransportMessageFrame, LivekitTransportMessageFrame


class LivekitOutputTransportProcessor(AudioCameraOutputProcessor):
    def __init__(self, client: LivekitTransportClient, params: LivekitParams, **kwargs):
        super().__init__(params, **kwargs)
        self._client = client

    async def start(self, frame: StartFrame):
        # Parent start.
        await super().start(frame)
        # Join the room.
        await self._client.join()

    async def stop(self, frame: EndFrame):
        # Parent stop.
        await super().stop(frame)
        # Leave the room.
        await self._client.leave()
        # Clear
        await self.cleanup()

    async def cancel(self, frame: CancelFrame):
        # Parent stop.
        await super().cancel(frame)
        # Leave the room.
        await self._client.leave()
        # Clear
        await self.cleanup()

    async def cleanup(self):
        await super().cleanup()
        await self._client.cleanup()

    async def send_message(self, frame: TransportMessageFrame):
        await self._client.send_message(frame)

    async def send_metrics(self, frame: MetricsFrame):
        metrics = {}
        if frame.ttfb:
            metrics["ttfb"] = frame.ttfb
        if frame.processing:
            metrics["processing"] = frame.processing
        if frame.tokens:
            metrics["tokens"] = frame.tokens
        if frame.characters:
            metrics["characters"] = frame.characters

        message = LivekitTransportMessageFrame(
            message={"type": "chatbot-metrics", "metrics": metrics}
        )
        await self._client.send_message(message)

    async def write_raw_audio_frames(self, frames: bytes):
        await self._client.write_raw_audio_frames(frames)

    async def write_frame_to_camera(self, frame: ImageRawFrame):
        await self._client.write_frame_to_camera(frame)



import os

os.environ["USE_LIBUV"] = "0"
import sys
from typing import Optional

import hydra
import lightning as L
import pyrootutils
import torch
from lightning import Callback, LightningDataModule, LightningModule, Trainer
from lightning.pytorch.loggers import Logger
from lightning.pytorch.strategies import DDPStrategy
from omegaconf import DictConfig, OmegaConf

os.environ.pop("SLURM_NTASKS", None)
os.environ.pop("SLURM_JOB_NAME", None)
os.environ.pop("SLURM_NTASKS_PER_NODE", None)

# register eval resolver and root
pyrootutils.setup_root(__file__, indicator=".project-root", pythonpath=True)

# Allow TF32 on Ampere GPUs
torch.set_float32_matmul_precision("high")
torch.backends.cudnn.allow_tf32 = True

# register eval resolver
OmegaConf.register_new_resolver("eval", eval)

import fish_speech.utils as utils

log = utils.RankedLogger(__name__, rank_zero_only=True)


@utils.task_wrapper
def train(cfg: DictConfig) -> tuple[dict, dict]:
    """Trains the model. Can additionally evaluate on a testset, using best weights obtained during
    training.
    This method is wrapped in optional @task_wrapper decorator, that controls the behavior during
    failure. Useful for multiruns, saving info about the crash, etc.
    Args:
        cfg (DictConfig): Configuration composed by Hydra.
    Returns:
        Tuple[dict, dict]: Dict with metrics and dict with all instantiated objects.
    """  # noqa: E501

    # set seed for random number generators in pytorch, numpy and python.random
    if cfg.get("seed"):
        L.seed_everything(cfg.seed, workers=False)

    if cfg.get("deterministic"):
        torch.use_deterministic_algorithms(True)

    log.info(f"Instantiating datamodule <{cfg.data._target_}>")
    datamodule: LightningDataModule = hydra.utils.instantiate(cfg.data)

    log.info(f"Instantiating model <{cfg.model._target_}>")
    model: LightningModule = hydra.utils.instantiate(cfg.model)

    log.info("Instantiating callbacks...")
    callbacks: list[Callback] = utils.instantiate_callbacks(cfg.get("callbacks"))

    log.info("Instantiating loggers...")
    logger: list[Logger] = utils.instantiate_loggers(cfg.get("logger"))

    log.info(f"Instantiating trainer <{cfg.trainer._target_}>")
    trainer: Trainer = hydra.utils.instantiate(
        cfg.trainer,
        callbacks=callbacks,
        logger=logger,
    )

    object_dict = {
        "cfg": cfg,
        "datamodule": datamodule,
        "model": model,
        "callbacks": callbacks,
        "logger": logger,
        "trainer": trainer,
    }

    if logger:
        log.info("Logging hyperparameters!")
        utils.log_hyperparameters(object_dict)

    if cfg.get("train"):
        log.info("Starting training!")

        ckpt_path = cfg.get("ckpt_path")
        auto_resume = False

        resume_ckpt_path = utils.get_latest_checkpoint(cfg.paths.ckpt_dir)
        if resume_ckpt_path is not None:
            ckpt_path = resume_ckpt_path
            auto_resume = True

        if ckpt_path is not None:
            log.info(f"Resuming from checkpoint: {ckpt_path}")

        # resume weights only is disabled for auto-resume
        if cfg.get("resume_weights_only") and auto_resume is False:
            log.info("Resuming weights only!")
            ckpt = torch.load(ckpt_path, map_location=model.device)
            if "state_dict" in ckpt:
                ckpt = ckpt["state_dict"]
            err = model.load_state_dict(ckpt, strict=False)
            log.info(f"Error loading state dict: {err}")
            ckpt_path = None

        trainer.fit(model=model, datamodule=datamodule, ckpt_path=ckpt_path)

    train_metrics = trainer.callback_metrics

    if cfg.get("test"):
        log.info("Starting testing!")
        ckpt_path = trainer.checkpoint_callback.best_model_path
        if ckpt_path == "":
            log.warning("Best ckpt not found! Using current weights for testing...")
            ckpt_path = cfg.get("ckpt_path")

        trainer.test(model=model, datamodule=datamodule, ckpt_path=ckpt_path)
        log.info(f"Best ckpt path: {ckpt_path}")

    test_metrics = trainer.callback_metrics

    # merge train and test metrics
    metric_dict = {**train_metrics, **test_metrics}

    return metric_dict, object_dict


@hydra.main(
    version_base="1.3", config_path="./configs", config_name="llama_pretrain.yaml"
)
def main(cfg: DictConfig) -> Optional[float]:
    # train the model
    train(cfg)


if __name__ == "__main__":
    main()



import os
import json
import argparse

def merge_outputs(out_path):
    files = os.listdir(out_path)
    # print(files)
    alist = []
    for file in files:
        alist += json.load(open(os.path.join(out_path, file), encoding='utf-8'))
    # print(len(alist))

    filename = out_path + "/results_final" + ".json"
    with open(filename, 'w', encoding="utf-8") as file_obj:
        json.dump(alist, file_obj, ensure_ascii=False, indent=1)

parser = argparse.ArgumentParser()
parser.add_argument("--out_path", type=str, required=True)
args = parser.parse_args()

merge_outputs(args.out_path)


"""Initialize funasr package."""

import logging
import os
import pkgutil
import importlib

try:
    from funasr_detach import version
    __version__ = version.__version__
except Exception as e:
    __version__ = "1.0.8"
    logging.warning(f"set __version__ to {__version__} because {e}")


import importlib
import pkgutil


def import_submodules(package, recursive=True):
    if isinstance(package, str):
        package = importlib.import_module(package)
    results = {}
    for loader, name, is_pkg in pkgutil.walk_packages(
        package.__path__, package.__name__ + "."
    ):
        try:
            results[name] = importlib.import_module(name)
        except Exception as e:
            # 如果想要看到导入错误的具体信息，可以取消注释下面的行
            # print(f"Failed to import {name}: {e}")
            pass
        if recursive and is_pkg:
            results.update(import_submodules(name))
    return results


import_submodules(__name__)

from funasr_detach.auto.auto_model import AutoModel
from funasr_detach.auto.auto_frontend import AutoFrontend



import torch
import torch.nn as nn


class MossFormerDecoder(nn.ConvTranspose1d):
    """A decoder layer that consists of ConvTranspose1d.

    Arguments
    ---------
    kernel_size : int
        Length of filters.
    in_channels : int
        Number of  input channels.
    out_channels : int
        Number of output channels.


    Example
    ---------
    >>> x = torch.randn(2, 100, 1000)
    >>> decoder = Decoder(kernel_size=4, in_channels=100, out_channels=1)
    >>> h = decoder(x)
    >>> h.shape
    torch.Size([2, 1003])
    """

    def __init__(self, *args, **kwargs):
        super(MossFormerDecoder, self).__init__(*args, **kwargs)

    def forward(self, x):
        """Return the decoded output.

        Arguments
        ---------
        x : torch.Tensor
            Input tensor with dimensionality [B, N, L].
                where, B = Batchsize,
                       N = number of filters
                       L = time points
        """

        if x.dim() not in [2, 3]:
            raise RuntimeError("{} accept 3/4D tensor as input".format(self.__name__))
        x = super().forward(x if x.dim() == 3 else torch.unsqueeze(x, 1))

        if torch.squeeze(x).dim() == 1:
            x = torch.squeeze(x, dim=1)
        else:
            x = torch.squeeze(x)
        return x



import os
import logging

from achatbot.common.http import HTTPRequest
from achatbot.common.factory import EngineClass, EngineFactory
from achatbot.common import interface
from achatbot.modules.functions.function import FunctionManager
from achatbot.common.types import SearchApiArgs, Search1ApiArgs, SerperApiArgs
import achatbot.modules.functions.search


class SearchBaseApi(EngineClass, interface.IFunction):
    def __init__(self) -> None:
        self.requests = HTTPRequest()

    def get_tool_call(self):
        return {
            "type": "function",
            "function": {
                "name": "web_search",
                "description": "web search by query",
                "parameters": {
                    "type": "object",
                    "properties": {"query": {"type": "string", "description": "web search query"}},
                    "required": ["query"],
                },
            },
        }

    def execute(self, session, **args):
        return self._web_search(session, **args)

    def _web_search(self, session, **args) -> str:
        pass


class SearchFuncEnvInit:
    engine: interface.IFunction | EngineClass = None

    @staticmethod
    def initSearchEngine() -> interface.IFunction | EngineClass:
        if SearchFuncEnvInit.engine is not None:
            logging.info(f"SearchFuncEnvInit.engine already initialized {SearchFuncEnvInit.engine}")
            return SearchFuncEnvInit.engine
        tag = os.getenv("FUNC_SEARCH_TAG", "search_api")
        kwargs = SearchFuncEnvInit.map_config_func[tag]()
        SearchFuncEnvInit.engine = EngineFactory.get_engine_by_tag(EngineClass, tag, **kwargs)
        logging.info(f"initSearchEngine: {tag}, {SearchFuncEnvInit.engine}")
        return SearchFuncEnvInit.engine

    @staticmethod
    def get_serper_api_args():
        return SerperApiArgs(
            gl=os.getenv("SERPER_GL", "cn"),
            hl=os.getenv("SERPER_HL", "zh-cn"),
            page=int(os.getenv("SERPER_PAGE", "1")),
            num=int(os.getenv("SERPER_NUM", "5")),
        ).__dict__

    @staticmethod
    def get_search_api_args():
        return SearchApiArgs(
            engine=os.getenv("SEARCH_ENGINE", "google"),
            gl=os.getenv("SEARCH_GL", "cn"),
            hl=os.getenv("SEARCH_HL", "zh-cn"),
            page=int(os.getenv("SERPER_PAGE", "1")),
            num=int(os.getenv("SERPER_NUM", "5")),
        ).__dict__

    @staticmethod
    def get_search1_api_args():
        return Search1ApiArgs(
            search_service=os.getenv("SEARCH1_ENGINE", "google"),
            image=bool(os.getenv("SEARCH1_IMAGE", "")),
            crawl_results=int(os.getenv("CRAWL_RESULTS", "0")),
            max_results=int(os.getenv("MAX_RESULTS", "5")),
        ).__dict__

    # TAG : config
    map_config_func = {
        "search_api": get_search_api_args,
        "search1_api": get_search1_api_args,
        "serper_api": get_serper_api_args,
    }


@FunctionManager.functions.register("web_search")
class SearchFunc:
    @staticmethod
    def get_tool_call():
        return SearchFuncEnvInit.initSearchEngine().get_tool_call()

    @staticmethod
    def execute(session, **args):
        return SearchFuncEnvInit.initSearchEngine().execute(session, **args)



r"""
use SOTA LLM like chatGPT to generate config file(json,yaml,toml) from dataclass type
"""

from dataclasses import dataclass, field
from enum import Enum
from typing import (
    Literal,
    Optional,
    Sequence,
    List,
    Any,
    Mapping,
    Union,
)
import os
from pathlib import Path

from dotenv import load_dotenv
from pydantic.main import BaseModel
from pydantic import ConfigDict

from .interface import (
    IBuffering,
    IDetector,
    IAsr,
    ILlm,
    ITts,
    IVADAnalyzer,
)
from .factory import EngineClass

load_dotenv(override=True)


SRC_PATH = os.path.normpath(os.path.join(os.path.dirname(os.path.abspath(__file__)), os.pardir))
DIR_PATH = os.path.join(SRC_PATH, os.pardir)
if bool(os.getenv("ACHATBOT_PKG", "")):
    home_dir = Path.home()
    DIR_PATH = os.path.join(home_dir, ".achatbot")

LOG_DIR = os.path.normpath(os.path.join(DIR_PATH, "log"))
CONFIG_DIR = os.path.normpath(os.path.join(DIR_PATH, "config"))
MODELS_DIR = os.path.normpath(os.path.join(DIR_PATH, "models"))
RECORDS_DIR = os.path.normpath(os.path.join(DIR_PATH, "records"))
VIDEOS_DIR = os.path.normpath(os.path.join(DIR_PATH, "videos"))
ASSETS_DIR = os.path.normpath(os.path.join(DIR_PATH, "assets"))

TEST_DIR = os.path.normpath(os.path.join(SRC_PATH, os.pardir, "test"))

# audio stream default configuration
CHUNK = 1600  # 100ms 16k rate num_frames
CHANNELS = 1
RATE = 16000
SAMPLE_WIDTH = 2

# llm sys default prompt
DEFAULT_SYSTEM_PROMPT = "你是一个中国人,一名中文助理，请用中文简短回答，回答限制在1-5句话内。要友好、乐于助人且简明扼要。保持对话简短而甜蜜。只用纯文本回答，不要包含链接或其他附加内容。不要回复计算机代码以及数学公式。"


@dataclass
class SessionCtx:
    client_id: str
    sampling_rate: int = RATE
    sample_width: int = SAMPLE_WIDTH
    read_audio_frames = bytes()
    state = dict()
    buffering_strategy: IBuffering = None
    waker: IDetector | EngineClass = None
    vad: IDetector | EngineClass = None
    asr: IAsr | EngineClass = None
    llm: ILlm | EngineClass = None
    tts: ITts | EngineClass = None
    on_session_start: callable = None
    on_session_end: callable = None

    def __repr__(self) -> str:
        d = {
            "client_id": self.client_id,
            "sampling_rate": self.sampling_rate,
            "sample_width": self.sample_width,
            "read_audio_frames_len": len(self.read_audio_frames),
            "state": self.state,
        }
        if "tts_chunk" in self.state:
            d["state"]["tts_chunk_len"] = len(self.state["tts_chunk"])
            d["state"].pop("tts_chunk")

        res = f"session ctx: {d}"
        return res

    def __getstate__(self):
        return {
            "client_id": self.client_id,
            "sampling_rate": self.sampling_rate,
            "sample_width": self.sample_width,
            "read_audio_frames": self.read_audio_frames,
            "state": self.state,
        }

    def __setstate__(self, state):
        self.client_id = state["client_id"]
        self.sampling_rate = state["sampling_rate"]
        self.sample_width = state["sample_width"]
        self.read_audio_frames = state["read_audio_frames"]
        self.state = state["state"]


# audio recorder default configuration
SILENCE_THRESHOLD = 500
# 1 seconds of silence marks the end of user voice input
SILENT_CHUNKS = 1 * RATE / CHUNK
# record silence timeout (10s)
SILENCE_TIMEOUT_S = 10
# Set microphone id. Use list_microphones.py to see a device list.
MIC_IDX = 1
# 2^(16-1)
INT16_MAX_ABS_VALUE = 32768.0

# pyaudio format
#   >>> print(int(pyaudio.paInt16))
#   8
PYAUDIO_PAINT16 = 8
#   >>> print(int(pyaudio.paInt24))
#   4
PYAUDIO_PAINT24 = 4
#   >>> print(int(pyaudio.paInt32))
#   2
PYAUDIO_PAINT32 = 2
#   >>> print(int(pyaudio.paFloat32))
#   1
PYAUDIO_PAFLOAT32 = 1
#   >>> print(int(pyaudio.paInt8))
#   16
PYAUDIO_PAINT8 = 16
#   >>> print(int(pyaudio.paUInt8))
#   32
PYAUDIO_PAUINT8 = 32
#   >>> print(int(pyaudio.paCustomFormat))
#   65536
PYAUDIO_PACUSTOMFORMAT = 65536

# PortAudio Callback Return Codes

PYAUDIO_PACONTINUE = 0  #: There is more audio data to come
PYAUDIO_PACOMPLETE = 1  #: This was the last block of audio data
PYAUDIO_PAABORT = 2  #: An error ocurred, stop playback/recording


@dataclass
class AudioStreamArgs:
    stream_callback: Optional[str] = None
    input: bool = False
    output: bool = False
    frames_per_buffer: int = CHUNK


@dataclass
class DailyAudioStreamArgs(AudioStreamArgs):
    """
    in live room,one joined client instance need diff in/out stream sample rate
    """

    meeting_room_url: str = ""
    bot_name: str = "chat-bot"
    meeting_room_token: str = ""
    in_channels: int = CHANNELS
    in_sample_rate: int = RATE
    in_sample_width: int = SAMPLE_WIDTH
    out_channels: int = CHANNELS
    out_sample_rate: int = RATE
    out_sample_width: int = SAMPLE_WIDTH
    out_queue_timeout_s: float = 0.1
    is_async_output: bool = False


@dataclass
class AudioRecoderArgs:
    silent_chunks: int = SILENT_CHUNKS
    silence_timeout_s: int = SILENCE_TIMEOUT_S
    num_frames: int = CHUNK
    is_stream_callback: bool = False
    no_stream_sleep_time_s: float = 0.03  # for dequeue get


@dataclass
class VADRecoderArgs(AudioRecoderArgs):
    padding_ms: int = 300
    active_ratio: float = 0.75
    silent_ratio: float = 0.75


@dataclass
class AudioPlayerArgs:
    on_play_start: Optional[str] = None
    on_play_end: Optional[str] = None
    on_play_chunk: Optional[str] = None
    is_immediate_stop: bool = False


@dataclass
class SilenceAtEndOfChunkArgs:
    chunk_length_seconds: float
    chunk_offset_seconds: float


VAD_CHECK_PER_FRAMES = 1
VAD_CHECK_ALL_FRAMES = 2


@dataclass
class WebRTCVADArgs:
    aggressiveness: int = 3  # 0,1,2,3
    sample_rate: int = RATE
    check_frames_mode: int = VAD_CHECK_PER_FRAMES
    frame_duration_ms: int = 20  # ms


@dataclass
class FSMNVADArgs:
    sample_rate: int = RATE
    model: str = "fsmn-vad"
    model_version: str = "v2.0.4"
    check_frames_mode: int = VAD_CHECK_PER_FRAMES
    # frame_duration_ms: int = 64  # ms,  frame_length: 1024 = 16000*64 / 1000


INIT_SILERO_SENSITIVITY = 0.4
# How often should we reset internal model state
SILERO_MODEL_RESET_STATES_TIME = 5.0


@dataclass
class SileroVADArgs:
    sample_rate: int = RATE
    repo_or_dir: str = "snakers4/silero-vad"
    model: str = "silero_vad"
    source: str = "github"  # github | local
    force_reload: bool = False
    trust_repo: bool = True
    verbose: bool = True
    onnx: bool = False
    silero_sensitivity: float = INIT_SILERO_SENSITIVITY
    is_pad_tensor: bool = False
    check_frames_mode: int = VAD_CHECK_PER_FRAMES


@dataclass
class WebRTCSileroVADArgs(WebRTCVADArgs, SileroVADArgs):
    pass


INIT_WAKE_WORDS_SENSITIVITY = 0.6
INIT_WAKE_WORD_ACTIVATION_DELAY = 0.0
INIT_WAKE_WORD_TIMEOUT = 5.0


@dataclass
class PorcupineDetectorArgs:
    keyword_paths: Optional[Sequence[str]] = None
    model_path: Optional[str] = None
    library_path: Optional[str] = None
    wake_words: str = ""
    wake_words_sensitivity: float = INIT_WAKE_WORDS_SENSITIVITY
    # wake_word_activation_delay: float = INIT_WAKE_WORD_ACTIVATION_DELAY
    # wake_word_timeout: float = INIT_WAKE_WORD_TIMEOUT
    on_wakeword_detected: Optional[str] = None
    # on_wakeword_timeout: Optional[str] = None
    # on_wakeword_detection_start: Optional[str] = None
    # on_wakeword_detection_end: Optional[str] = None


@dataclass
class LLamcppLLMArgs:
    save_chat_history: bool = True
    model_name: str = ""
    model_type: str = "generate"  # generate < chat | chat-func
    model_path: str = ""
    n_threads: int = 1
    n_batch: int = 8
    n_gpu_layers: int = 0
    n_ctx: int = 4096
    # chatml | chatml-function-calling | functionary-v2 | minicpm-v-2.6 ..
    chat_format: Optional[str] = None
    tokenizer_path: Optional[str] = None
    clip_model_path: Optional[str] = None
    verbose: bool = True
    flash_attn: bool = False
    # llm
    llm_prompt_tpl: str = "<|user|>\n{%s}<|end|>\n<|assistant|>"
    llm_stop: Optional[Union[int, List[str]]] = field(default_factory=list)
    llm_stop_ids: Optional[Union[int, List[int]]] = field(default_factory=list)
    llm_max_tokens: int = 2048
    llm_temperature: float = 0.8
    llm_top_p: float = 0.95
    llm_min_p: float = 0.05
    llm_top_k: int = 40
    llm_seed: Optional[int] = None
    # repeat penalty
    llm_repeat_penalty: float = 1.1
    llm_repeat_last_n: int = 64
    # chat stream
    llm_stream: bool = False
    llm_chat_system: str = ""
    # "none" | "auto" | dict function like this: { "type": "function", "function": { "name": "get_current_weather" } },
    llm_tool_choice: str = None


@dataclass
class PersonalAIProxyArgs:
    max_retry_cn: int = 2
    api_url: str = "http://localhost:8787/"
    chat_bot: str = "openai"  # openai | qianfan
    openai_api_base_url: str = "https://api.groq.com/openai/v1/"
    model_type: str = "chat_only"  # chat_only | chat_with_functions
    model_name: str = ""
    llm_chat_system: str = ""
    llm_max_tokens: int = 1024
    llm_stream: bool = False
    llm_stop: Optional[List[str]] = None
    llm_temperature: float = 0.8
    llm_top_p: float = 0.95
    func_search_name: str = "search_api"
    func_weather_name: str = "openweathermap_api"


# ---------------- TTS -----------------
@dataclass
class CoquiTTSArgs:
    conf_file: str = ""
    model_path: str = ""
    reference_audio_path: str = ""
    tts_temperature: float = 0.75
    tts_top_p: float = 0.85
    tts_stream: bool = False
    tts_length_penalty: float = 1.0
    tts_repetition_penalty: float = 10.0
    tts_num_beams: int = 1
    tts_speed: float = 1.0
    tts_stream_chunk_size: int = 20
    tts_overlap_wav_len: int = 1024
    tts_enable_text_splitting: bool = False
    tts_language: str = "zh"
    tts_default_silence_duration: float = 0.3
    tts_comma_silence_duration: float = 0.3
    tts_sentence_silence_duration: float = 0.6
    tts_use_deepspeed: bool = False


PYTTSX3_SYNTHESIS_FILE = "pyttsx3_synthesis.wav"


@dataclass
class Pyttsx3TTSArgs:
    voice_name: str = "Tingting"


GTTS_SYNTHESIS_FILE = "gtts_synthesis.wav"


@dataclass
class GTTSArgs:
    language: str = "en"
    tld: str = "com"
    slow: bool = False
    speed_increase: float = 1.0
    chunk_size: int = 100
    crossfade_lenght: int = 10


EDGE_TTS_SYNTHESIS_FILE = "edge_tts_synthesis.wav"


@dataclass
class EdgeTTSArgs:
    language: str = "en"
    gender: str = "Female"
    voice_name: str = "en-GB-SoniaNeural"
    rate: str = "+15%"
    volume: str = "+0%"
    pitch: str = "+0Hz"


@dataclass
class CosyVoiceTTSArgs:
    r"""
    CosyVoice:
    For zero_shot/cross_lingual/vc(voice convert) inference, please use CosyVoice-300M model.
    For sft inference, please use CosyVoice-300M-SFT model.
    For instruct inference, please use CosyVoice-300M-Instruct model.

    CosyVoice2:
    - if no reference_text and no instruct_text, default use cross lingual infer;
        fine grained control, for supported control, check cosyvoice/tokenizer/tokenizer.py#L248
    - if have reference_text and no instruct_text, use zero shot infer;
    - if have no reference_text and have instruct_text, use instruct2 infer;
    """

    model_dir: str = os.path.join(MODELS_DIR, "CosyVoice-300M-SFT")  # sft model

    reference_text: str = ""  # reference audio text
    reference_audio_path: str = ""  # 16k sample rate audio file for base pt model

    src_audio_path: str = ""  # for voice convert from src to reference audio

    instruct_text: str = ""  # use with instruct model
    spk_id: str = ""  # use with sft and instruct model when use CosyVoice, CosyVoice2 instruct2 method remove spk_id

    language: str = "zh"
    tts_default_silence_duration: float = 0.3
    tts_comma_silence_duration: float = 0.3
    tts_sentence_silence_duration: float = 0.6
    tts_speed: float = 1.0

    # stream
    tts_stream: bool = False
    chunk_length_seconds: float = 1.0


# ------------- llm function calling -----------------
@dataclass
class OpenWeatherMapArgs:
    lang: str = "zh_cn"
    units: str = "metric"


@dataclass
class SearchApiArgs:
    engine: str = "google"
    gl: str = "cn"
    hl: str = "zh-cn"
    page: int = 1
    num: int = 5


@dataclass
class Search1ApiArgs:
    search_service: str = "google"
    image: bool = False
    crawl_results: int = 0
    max_results: int = 5


@dataclass
class SerperApiArgs:
    gl: str = "cn"
    hl: str = "zh-cn"
    page: int = 1
    num: int = 5


# --------------- vad analyzer------------------
DAILY_WEBRTC_VAD_RESET_PERIOD_MS = 2000


class VADState(Enum):
    QUIET = 1
    STARTING = 2
    SPEAKING = 3
    STOPPING = 4


@dataclass
class VADAnalyzerArgs:
    sample_rate: int = RATE
    num_channels: int = CHANNELS
    confidence: float = 0.7
    start_secs: float = 0.2
    stop_secs: float = 0.8
    min_volume: float = 0.6


@dataclass
class SileroVADAnalyzerArgs(SileroVADArgs, VADAnalyzerArgs):
    pass


# --------------- in/out audio camera(video) params -------------------


class AudioParams(BaseModel):
    audio_out_enabled: bool = False
    audio_out_sample_rate: int = RATE
    audio_out_channels: int = CHANNELS
    audio_in_enabled: bool = False
    audio_in_participant_enabled: bool = False
    audio_in_sample_rate: int = RATE
    audio_in_channels: int = CHANNELS


class AudioVADParams(AudioParams):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    vad_enabled: bool = False
    vad_audio_passthrough: bool = False
    vad_analyzer: IVADAnalyzer | EngineClass | None = None


class CameraParams(BaseModel):
    camera_in_enabled: bool = False
    camera_in_color_format: str = "RGB"
    camera_out_enabled: bool = False
    camera_out_is_live: bool = False
    camera_out_width: int = 1024
    camera_out_height: int = 768
    camera_out_bitrate: int = 800000
    camera_out_framerate: int = 30
    camera_out_color_format: str = "RGB"


class AudioCameraParams(CameraParams, AudioVADParams):
    model_config = ConfigDict(arbitrary_types_allowed=True)


# --------------- daily -------------------------------


class DailyDialinSettings(BaseModel):
    call_id: str = ""
    call_domain: str = ""


# https://reference-python.daily.co/types.html#transcriptionsettings
class DailyTranscriptionSettings(BaseModel):
    language: str = "en"
    tier: str = "nova"
    model: str = "2-conversationalai"
    profanity_filter: bool = True
    redact: bool = False
    endpointing: bool = True
    punctuate: bool = True
    includeRawResponse: bool = True
    extra: Mapping[str, Any] = {"interim_results": True}


class DailyParams(AudioCameraParams):
    api_url: str = "https://api.daily.co/v1"
    api_key: str = ""
    dialin_settings: DailyDialinSettings | None = None
    transcription_enabled: bool = False
    transcription_settings: DailyTranscriptionSettings = DailyTranscriptionSettings()


class DailyRoomArgs(BaseModel):
    privacy: Literal["private", "public"] = "public"


# --------------- livekit -------------------------------


class LivekitParams(AudioCameraParams):
    # audio_in_sample_rate: int = 48000  # livekit audio in stream default sample rate 48000
    websocket_url: str = ""  # project url
    api_key: str = ""
    api_secret: str = ""
    e2ee_shared_key: Optional[bytes] = None
    sandbox_room_url: str = "https://ultra-terminal-re8nmd.sandbox.livekit.io"


class LivekitRoomArgs(BaseModel):
    bot_name: str = "chat-bot"
    # if use session, need manual close async http session
    is_common_session: bool = False


# --------------- agora -------------------------------


class AgoraParams(AudioCameraParams):
    app_id: str = ""
    app_cert: str = ""
    enable_pcm_dump: bool = False
    demo_voice_url: str = "https://webdemo.agora.io/basicVoiceCall/index.html"
    demo_video_url: str = "https://webdemo.agora.io/basicVideoCall/index.html"


class AgoraChannelArgs(BaseModel):
    pass


# ---------------- Room Bots -------------
class GeneralRoomInfo(BaseModel):
    """general room info for diff webRTC room info to align"""

    sid: str = ""
    name: str = ""
    url: str = ""
    ttl_s: int | None = None
    creation_time: str | None = None
    extra_data: dict = field(default_factory=dict)


@dataclass
class BotRunArgs:
    room_name: str = ""
    room_url: str = ""
    token: str = ""
    bot_config: dict | None = None
    bot_name: str | None = None
    bot_config_list: list | None = None
    services: dict | None = None
    websocket_server_host: str = "localhost"
    websocket_server_port: int = 8765
    handle_sigint: bool = True
